import unittest

from django import forms
from django.contrib.auth.models import User
from django.http import QueryDict
from kgb import SpyAgency

from reviewboard.hostingsvcs.base import hosting_service_registry
from reviewboard.hostingsvcs.models import HostingServiceAccount
from reviewboard.hostingsvcs.github import GitHub
from reviewboard.scmtools import scmtools_registry
from reviewboard.scmtools.certs import Certificate
from reviewboard.scmtools.errors import UnverifiedCertificateError
from reviewboard.scmtools.forms import (RepositoryForm,
                                        StandardSCMToolAuthForm,
                                        StandardSCMToolRepositoryForm)
from reviewboard.scmtools.git import GitTool
from reviewboard.scmtools.models import Tool
from reviewboard.scmtools.perforce import PerforceTool
from reviewboard.site.models import LocalSite
from reviewboard.testing.hosting_services import (SelfHostedTestService,
                                                  TestService)
from reviewboard.testing.testcase import TestCase

try:
    import P4
except ImportError:
    P4 = None


class HiddenTestService(TestService):
    hosting_service_id = 'hidden-test'
    name = 'Hidden Test Service'
    visible = False


class RepositoryFormTests(SpyAgency, TestCase):
    """Unit tests for the repository form."""

    fixtures = ['test_scmtools']

    def setUp(self):
        super(RepositoryFormTests, self).setUp()

        hosting_service_registry.register(TestService)
        hosting_service_registry.register(SelfHostedTestService)
        hosting_service_registry.register(HiddenTestService)

    def tearDown(self):
        super(RepositoryFormTests, self).tearDown()

        hosting_service_registry.unregister(TestService)
        hosting_service_registry.unregister(SelfHostedTestService)
        hosting_service_registry.unregister(HiddenTestService)

    def test_without_localsite(self):
        """Testing RepositoryForm without a LocalSite"""
        local_site = LocalSite.objects.create(name='test')
        local_site_user = User.objects.create_user(username='testuser1')
        local_site.users.add(local_site_user)

        global_site_user = User.objects.create_user(username='testuser2')

        local_site_group = self.create_review_group(name='test1',
                                                    invite_only=True,
                                                    local_site=local_site)
        global_site_group = self.create_review_group(name='test2',
                                                     invite_only=True)

        local_site_account = HostingServiceAccount.objects.create(
            username='local-test-user',
            service_name='test',
            local_site=local_site)
        global_site_account = HostingServiceAccount.objects.create(
            username='global-test-user',
            service_name='test')

        # Make sure the initial state and querysets are what we expect on init.
        form = RepositoryForm()

        self.assertIsNone(form.limited_to_local_site)
        self.assertIn('local_site', form.fields)
        self.assertEqual(list(form.fields['users'].queryset),
                         [local_site_user, global_site_user])
        self.assertEqual(list(form.fields['review_groups'].queryset),
                         [local_site_group, global_site_group])
        self.assertEqual(list(form.fields['hosting_account'].queryset),
                         [local_site_account, global_site_account])

        # Now test what happens when it's been fed data and validated.
        form = self._build_form({
            'name': 'test',
            'tool': 'git',
            'path': '/path/to/test.git',
            'mirror_path': 'git@localhost:test.git',
            'public': False,
            'users': [global_site_user.pk],
            'review_groups': [global_site_group.pk],
        })

        self.assertIsNone(form.limited_to_local_site)
        self.assertIn('local_site', form.fields)
        self.assertEqual(list(form.fields['users'].queryset),
                         [local_site_user, global_site_user])
        self.assertEqual(list(form.fields['review_groups'].queryset),
                         [local_site_group, global_site_group])
        self.assertIsNone(form.fields['users'].widget.local_site_name)
        self.assertEqual(list(form.fields['hosting_account'].queryset),
                         [local_site_account, global_site_account])
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.scmtool_repository_forms['git'],
            ])

        self.assertTrue(form.is_valid())

        # Make sure any overridden querysets have been restored, so users can
        # still change entries.
        self.assertEqual(list(form.fields['users'].queryset),
                         [local_site_user, global_site_user])
        self.assertEqual(list(form.fields['review_groups'].queryset),
                         [local_site_group, global_site_group])
        self.assertEqual(list(form.fields['hosting_account'].queryset),
                         [local_site_account, global_site_account])

        repository = form.save()
        form.save_m2m()

        self.assertEqual(repository.scmtool_id, 'git')
        self.assertIsNone(repository.local_site)
        self.assertEqual(repository.extra_data, {})
        self.assertEqual(list(repository.users.all()), [global_site_user])
        self.assertEqual(list(repository.review_groups.all()),
                         [global_site_group])

        self.assertSpyCalledWith(GitTool.check_repository,
                                 path='/path/to/test.git',
                                 username='',
                                 password='',
                                 local_site_name=None)

    def test_without_localsite_and_instance(self):
        """Testing RepositoryForm without a LocalSite and editing instance"""
        local_site = LocalSite.objects.create(name='test')
        repository = self.create_repository(local_site=local_site)

        form = self._build_form(
            data={
                'name': 'test',
                'tool': 'git',
                'path': '/path/to/test.git',
                'mirror_path': 'git@localhost:test.git',
            },
            instance=repository)
        self.assertEqual(form.fields['tool'].initial, 'git')

        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['tool'],
                         scmtools_registry.get_by_id('git'))
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.scmtool_repository_forms['git'],
            ])

        new_repository = form.save()
        self.assertEqual(repository.pk, new_repository.pk)
        self.assertEqual(repository.extra_data, {})
        self.assertEqual(repository.path, '/path/to/test.git')
        self.assertEqual(repository.mirror_path, 'git@localhost:test.git')
        self.assertIsNone(new_repository.local_site)
        self.assertEqual(new_repository.tool,
                         Tool.objects.get(name='Git'))
        self.assertEqual(new_repository.scmtool_id, 'git')

        self.assertSpyCalledWith(GitTool.check_repository,
                                 path='/path/to/test.git',
                                 username='',
                                 password='',
                                 local_site_name=None)

    def test_without_localsite_and_with_local_site_user(self):
        """Testing RepositoryForm without a LocalSite and User on a LocalSite
        """
        local_site = LocalSite.objects.create(name='test')
        user = User.objects.create_user(username='testuser1')
        local_site.users.add(user)

        form = self._build_form({
            'name': 'test',
            'tool': 'git',
            'path': '/path/to/test.git',
            'users': [user.pk],
        })
        self.assertTrue(form.is_valid())
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.scmtool_repository_forms['git'],
            ])

    def test_without_localsite_and_with_local_site_group(self):
        """Testing RepositoryForm without a LocalSite and Group on a LocalSite
        """
        local_site = LocalSite.objects.create(name='test')
        group = self.create_review_group(local_site=local_site)

        form = self._build_form({
            'name': 'test',
            'tool': 'git',
            'path': '/path/to/test.git',
            'review_groups': [group.pk],
        })
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'review_groups': [
                    'Select a valid choice. 1 is not one of the available '
                    'choices.',
                ],
            })
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.scmtool_repository_forms['git'],
            ])

    def test_without_localsite_and_with_local_site_hosting_account(self):
        """Testing RepositoryForm without a LocalSite and
        HostingServiceAccount on a LocalSite
        """
        local_site = LocalSite.objects.create(name='test')

        hosting_account = HostingServiceAccount.objects.create(
            username='test-user',
            service_name='test',
            local_site=local_site)

        form = self._build_form({
            'name': 'test',
            'hosting_type': 'test',
            'hosting_account': hosting_account.pk,
            'test_repo_name': 'test',
            'tool': 'git',
        })
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'hosting_account': [
                    'Select a valid choice. That choice is not one of the '
                    'available choices.',
                ],
            })
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['test']['default'],
            ])

    def test_with_limited_localsite(self):
        """Testing RepositoryForm limited to a LocalSite"""
        local_site = LocalSite.objects.create(name='test')
        local_site_user = User.objects.create_user(username='testuser1')
        local_site.users.add(local_site_user)

        User.objects.create_user(username='testuser2')

        local_site_group = self.create_review_group(name='test1',
                                                    invite_only=True,
                                                    local_site=local_site)
        self.create_review_group(name='test2', invite_only=True)

        form = self._build_form(limit_to_local_site=local_site)

        self.assertEqual(form.limited_to_local_site, local_site)
        self.assertNotIn('local_site', form.fields)
        self.assertEqual(list(form.fields['users'].queryset),
                         [local_site_user])
        self.assertEqual(list(form.fields['review_groups'].queryset),
                         [local_site_group])
        self.assertEqual(form.fields['users'].widget.local_site_name,
                         local_site.name)
        self.assertEqual(list(form.iter_subforms(bound_only=True)), [])

    def test_with_limited_localsite_and_changing_site(self):
        """Testing RepositoryForm limited to a LocalSite and changing
        LocalSite
        """
        local_site1 = LocalSite.objects.create(name='test-site-1')
        local_site2 = LocalSite.objects.create(name='test-site-2')

        form = self._build_form(
            {
                'name': 'test',
                'tool': 'git',
                'path': '/path/to/test.git',
                'local_site': local_site2.pk,
            },
            limit_to_local_site=local_site1)

        self.assertEqual(form.limited_to_local_site, local_site1)
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['local_site'], local_site1)
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.scmtool_repository_forms['git'],
            ])

        repository = form.save()
        self.assertEqual(repository.local_site, local_site1)

        self.assertSpyCalledWith(GitTool.check_repository,
                                 path='/path/to/test.git',
                                 username='',
                                 password='',
                                 local_site_name='test-site-1')

    def test_with_limited_localsite_and_compatible_instance(self):
        """Testing RepositoryForm limited to a LocalSite and editing compatible
        instance
        """
        local_site = LocalSite.objects.create(name='test')
        repository = self.create_repository(local_site=local_site)

        # This should just simply not raise an exception.
        RepositoryForm(instance=repository,
                       limit_to_local_site=local_site)

    def test_with_limited_localsite_and_incompatible_instance(self):
        """Testing RepositoryForm limited to a LocalSite and editing
        incompatible instance
        """
        local_site = LocalSite.objects.create(name='test')
        repository = self.create_repository()

        error_message = (
            'The provided instance is not associated with a LocalSite '
            'compatible with this form. Please contact support.'
        )

        with self.assertRaisesMessage(ValueError, error_message):
            RepositoryForm(instance=repository,
                           limit_to_local_site=local_site)

    def test_with_limited_localsite_and_invalid_user(self):
        """Testing DefaultReviewerForm limited to a LocalSite with a User
        not on the LocalSite
        """
        local_site = LocalSite.objects.create(name='test')
        user = User.objects.create_user(username='test')

        form = self._build_form(
            {
                'name': 'test',
                'tool': 'git',
                'path': '/path/to/test.git',
                'users': [user.pk],
            },
            limit_to_local_site=local_site)

        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'users': [
                    'Select a valid choice. 1 is not one of the available '
                    'choices.',
                ],
            })
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.scmtool_repository_forms['git'],
            ])

    def test_with_limited_localsite_and_invalid_group(self):
        """Testing DefaultReviewerForm limited to a LocalSite with a Group
        not on the LocalSite
        """
        local_site = LocalSite.objects.create(name='test')
        group = self.create_review_group()

        form = self._build_form(
            {
                'name': 'test',
                'tool': 'git',
                'path': '/path/to/test.git',
                'review_groups': [group.pk],
            },
            limit_to_local_site=local_site)

        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'review_groups': [
                    'Select a valid choice. 1 is not one of the available '
                    'choices.',
                ],
            })
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.scmtool_repository_forms['git'],
            ])

    def test_with_limited_localsite_and_invalid_hosting_account(self):
        """Testing DefaultReviewerForm limited to a LocalSite with a
        HostingServiceAccount not on the LocalSite
        """
        local_site = LocalSite.objects.create(name='test')

        hosting_account = HostingServiceAccount.objects.create(
            username='test-user',
            service_name='test')

        form = self._build_form(
            {
                'name': 'test',
                'hosting_type': 'test',
                'hosting_account': hosting_account.pk,
                'test_repo_name': 'test',
                'tool': 'git',
            },
            limit_to_local_site=local_site)

        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'hosting_account': [
                    'Select a valid choice. That choice is not one of the '
                    'available choices.',
                ],
            })
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['test']['default'],
            ])

    def test_with_localsite_in_data(self):
        """Testing RepositoryForm with a LocalSite in form data"""
        local_site = LocalSite.objects.create(name='test')
        local_site_user = User.objects.create_user(username='testuser1')
        local_site.users.add(local_site_user)

        global_site_user = User.objects.create_user(username='testuser2')

        local_site_group = self.create_review_group(name='test1',
                                                    invite_only=True,
                                                    local_site=local_site)
        global_site_group = self.create_review_group(name='test2',
                                                     invite_only=True)

        local_site_account = HostingServiceAccount.objects.create(
            username='local-test-user',
            service_name='test',
            local_site=local_site)
        local_site_account.data['password'] = 'testpass'
        local_site_account.save(update_fields=('data',))

        global_site_account = HostingServiceAccount.objects.create(
            username='global-test-user',
            service_name='test')

        # Make sure the initial state and querysets are what we expect on init.
        form = RepositoryForm()

        self.assertIsNone(form.limited_to_local_site)
        self.assertIn('local_site', form.fields)
        self.assertEqual(list(form.fields['users'].queryset),
                         [local_site_user, global_site_user])
        self.assertEqual(list(form.fields['review_groups'].queryset),
                         [local_site_group, global_site_group])
        self.assertIsNone(form.fields['users'].widget.local_site_name)
        self.assertEqual(list(form.fields['hosting_account'].queryset),
                         [local_site_account, global_site_account])

        # Now test what happens when it's been fed data and validated.
        form = self._build_form({
            'name': 'test',
            'hosting_type': 'test',
            'hosting_account': local_site_account.pk,
            'test_repo_name': 'test',
            'tool': 'git',
            'path': '/path/to/test.git',
            'public': False,
            'local_site': local_site.pk,
            'users': [local_site_user.pk],
            'review_groups': [local_site_group.pk],
        })

        self.assertIsNone(form.limited_to_local_site)
        self.assertIn('local_site', form.fields)
        self.assertEqual(list(form.fields['users'].queryset),
                         [local_site_user, global_site_user])
        self.assertEqual(list(form.fields['review_groups'].queryset),
                         [local_site_group, global_site_group])
        self.assertIsNone(form.fields['users'].widget.local_site_name)
        self.assertEqual(list(form.fields['hosting_account'].queryset),
                         [local_site_account, global_site_account])
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['test']['default'],
            ])

        self.assertTrue(form.is_valid())

        # Make sure any overridden querysets have been restored, so users can
        # still change entries.
        self.assertEqual(list(form.fields['users'].queryset),
                         [local_site_user, global_site_user])
        self.assertEqual(list(form.fields['review_groups'].queryset),
                         [local_site_group, global_site_group])
        self.assertEqual(list(form.fields['hosting_account'].queryset),
                         [local_site_account, global_site_account])

        repository = form.save()
        form.save_m2m()

        self.assertEqual(repository.scmtool_id, 'git')
        self.assertEqual(repository.local_site, local_site)
        self.assertEqual(repository.hosting_account, local_site_account)
        self.assertEqual(
            repository.extra_data,
            {
                'bug_tracker_use_hosting': False,
                'repository_plan': '',
                'test_repo_name': 'test',
            })
        self.assertEqual(list(repository.users.all()), [local_site_user])
        self.assertEqual(list(repository.review_groups.all()),
                         [local_site_group])

        self.assertSpyCalledWith(TestService.check_repository,
                                 path='http://example.com/test/',
                                 username=None,
                                 password=None,
                                 local_site_name='test',
                                 plan='default',
                                 scmtool_class=GitTool,
                                 tool_name='Git')

    def test_with_localsite_in_data_and_instance(self):
        """Testing RepositoryForm with a LocalSite in form data and editing
        instance
        """
        local_site = LocalSite.objects.create(name='test')
        repository = self.create_repository()

        form = self._build_form(
            {
                'name': 'test',
                'tool': 'git',
                'path': '/path/to/test.git',
                'local_site': local_site.pk,
            },
            instance=repository)
        self.assertEqual(form.fields['tool'].initial, 'git')

        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['tool'],
                         scmtools_registry.get_by_id('git'))
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.scmtool_repository_forms['git'],
            ])

        new_repository = form.save()
        self.assertEqual(repository.pk, new_repository.pk)
        self.assertEqual(new_repository.local_site, local_site)
        self.assertEqual(new_repository.tool,
                         Tool.objects.get(name='Git'))
        self.assertEqual(new_repository.scmtool_id, 'git')
        self.assertEqual(repository.extra_data, {})

        self.assertSpyCalledWith(GitTool.check_repository,
                                 path='/path/to/test.git',
                                 username='',
                                 password='',
                                 local_site_name='test')

    def test_with_localsite_in_data_and_invalid_user(self):
        """Testing RepositoryForm with a LocalSite in form data and User not
        on the LocalSite
        """
        local_site = LocalSite.objects.create(name='test')
        user = User.objects.create_user(username='test-user')

        form = self._build_form({
            'name': 'test',
            'tool': 'git',
            'path': '/path/to/test.git',
            'local_site': local_site.pk,
            'users': [user.pk],
        })
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'users': [
                    'Select a valid choice. 1 is not one of the available '
                    'choices.',
                ],
            })
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.scmtool_repository_forms['git'],
            ])

    def test_with_localsite_in_data_and_invalid_group(self):
        """Testing RepositoryForm with a LocalSite in form data and Group not
        on the LocalSite
        """
        local_site = LocalSite.objects.create(name='test')
        group = self.create_review_group()

        form = self._build_form({
            'name': 'test',
            'tool': 'git',
            'path': '/path/to/test.git',
            'local_site': local_site.pk,
            'review_groups': [group.pk],
        })
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'review_groups': [
                    'Select a valid choice. 1 is not one of the available '
                    'choices.',
                ],
            })
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.scmtool_repository_forms['git'],
            ])

    def test_with_instance_and_no_hosting_service(self):
        """Testing RepositoryForm with instance= and no hosting service"""
        repository = self.create_repository(name='Test Repository',
                                            path='/path/',
                                            mirror_path='/mirror_path/',
                                            raw_file_url='/raw_file/',
                                            username='test-user',
                                            password='test-pass',
                                            encoding='utf-128',
                                            tool_name='Test')

        form = RepositoryForm(instance=repository)
        self.assertEqual(form['encoding'].value(), 'utf-128')
        self.assertEqual(form['name'].value(), 'Test Repository')
        self.assertEqual(form['tool'].value(), 'test')

        auth_form = form.scmtool_auth_forms['test']
        self.assertEqual(auth_form['username'].value(), 'test-user')
        self.assertEqual(auth_form['password'].value(), 'test-pass')

        scmtool_form = form.scmtool_repository_forms['test']
        self.assertEqual(scmtool_form['mirror_path'].value(), '/mirror_path/')
        self.assertEqual(scmtool_form['path'].value(), '/path/')
        self.assertEqual(scmtool_form['raw_file_url'].value(), '/raw_file/')

        # Defaults.
        self.assertEqual(form['bug_tracker'].value(), '')
        self.assertEqual(form['bug_tracker_type'].value(),
                         form.NO_BUG_TRACKER_ID)
        self.assertEqual(form['hosting_type'].value(),
                         form.NO_HOSTING_SERVICE_ID)
        self.assertIsNone(form['bug_tracker_hosting_account_username'].value())
        self.assertIsNone(form['bug_tracker_hosting_url'].value())
        self.assertIsNone(form['bug_tracker_plan'].value())
        self.assertIsNone(form['hosting_account'].value())
        self.assertIsNone(form['repository_plan'].value())
        self.assertTrue(form['public'].value())
        self.assertTrue(form['visible'].value())
        self.assertFalse(form['associate_ssh_key'].value())
        self.assertFalse(form['bug_tracker_use_hosting'].value())
        self.assertFalse(form['force_authorize'].value())
        self.assertFalse(form['reedit_repository'].value())
        self.assertFalse(form['trust_host'].value())

        # Check a couple more forms to make sure their data wasn't loaded.
        auth_form = form.scmtool_auth_forms['git']
        self.assertIsNone(auth_form['username'].value())
        self.assertIsNone(auth_form['password'].value())

        scmtool_form = form.scmtool_repository_forms['git']
        self.assertIsNone(scmtool_form['mirror_path'].value())
        self.assertIsNone(scmtool_form['path'].value())
        self.assertIsNone(scmtool_form['raw_file_url'].value())

    def test_with_instance_and_hosting_service(self):
        """Testing RepositoryForm with instance= and hosting service"""
        account = HostingServiceAccount.objects.create(username='test-user',
                                                       service_name='github')
        account.data['password'] = 'test-pass'
        account.save()

        repository = self.create_repository(name='Test Repository',
                                            path='/path/',
                                            mirror_path='/mirror_path/',
                                            raw_file_url='/raw_file/',
                                            encoding='utf-128',
                                            tool_name='Git',
                                            hosting_account=account)
        repository.extra_data.update({
            'bug_tracker_hosting_url': 'http://example.com/',
            'bug_tracker_type': 'github',
            'bug_tracker-hosting_account_username': 'test-user',
            'bug_tracker-github_repo_name': 'test-repo',
            'bug_tracker_plan': 'private',
            'repository_plan': 'public',
        })

        form = RepositoryForm(instance=repository)
        self.assertEqual(form['bug_tracker_hosting_account_username'].value(),
                         'test-user')
        self.assertEqual(form['bug_tracker_hosting_url'].value(),
                         'http://example.com/')
        self.assertEqual(form['bug_tracker_plan'].value(), 'private')
        self.assertEqual(form['bug_tracker_type'].value(), 'github')
        self.assertEqual(form['encoding'].value(), 'utf-128')
        self.assertEqual(form['hosting_account'].value(), account.pk)
        self.assertEqual(form['hosting_type'].value(), 'github')
        self.assertEqual(form['name'].value(), 'Test Repository')
        self.assertEqual(form['repository_plan'].value(), 'public')
        self.assertEqual(form['tool'].value(), 'git')

        # Since we used a hosting service above, the SCMTool form should not
        # receive any values (even though tool=git).
        scmtool_form = form.scmtool_repository_forms['git']
        self.assertIsNone(scmtool_form['mirror_path'].value())
        self.assertIsNone(scmtool_form['path'].value())
        self.assertIsNone(scmtool_form['raw_file_url'].value())

        # Defaults.
        self.assertTrue(form['visible'].value())
        self.assertTrue(form['public'].value())
        self.assertFalse(form['reedit_repository'].value())
        self.assertFalse(form['trust_host'].value())
        self.assertFalse(form['force_authorize'].value())
        self.assertFalse(form['associate_ssh_key'].value())
        self.assertFalse(form['bug_tracker_use_hosting'].value())

    def test_with_instance_and_public_and_acl(self):
        """Testing RepositoryForm with instance= and access lists set"""
        repository = self.create_repository(tool_name='Test',
                                            public=True)
        repository.users.add(self.create_user())
        repository.review_groups.add(self.create_review_group())

        form = RepositoryForm(instance=repository)
        self.assertEqual(form['users'].value(), [])
        self.assertEqual(form['review_groups'].value(), [])

    def test_plain_repository(self):
        """Testing RepositoryForm with a plain repository"""
        form = self._build_form({
            'name': 'test',
            'tool': 'git',
            'path': '/path/to/test.git',
            'mirror_path': 'git@localhost:test.git',
            'raw_file_url':
                'https://git.example.com/raw/<revision>/<filename>/',
            'username': 'myuser',
            'password': 'mypass',
        })

        self.assertTrue(form.is_valid())

        repository = form.save()
        self.assertEqual(repository.name, 'test')
        self.assertEqual(repository.tool, Tool.objects.get(name='Git'))
        self.assertEqual(repository.scmtool_id, 'git')
        self.assertEqual(repository.hosting_account, None)
        self.assertEqual(repository.path, '/path/to/test.git')
        self.assertEqual(repository.mirror_path, 'git@localhost:test.git')
        self.assertEqual(repository.raw_file_url,
                         'https://git.example.com/raw/<revision>/<filename>/')
        self.assertEqual(repository.extra_data, {})
        self.assertEqual(repository.get_credentials(), {
            'username': 'myuser',
            'password': 'mypass',
        })
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.scmtool_repository_forms['git'],
            ])

        # Make sure none of the other auth forms are unhappy. That would be
        # an indicator that we're doing form processing and validation wrong.
        for auth_form in form.hosting_auth_forms.values():
            self.assertEqual(auth_form.errors, {})

    def test_plain_repository_with_prefixed_standard_fields(self):
        """Testing RepositoryForm with a plain repository and prefixed
        standard fields
        """
        local_site = LocalSite.objects.create(name='test')
        repository = self.create_repository(local_site=local_site)

        form = self._build_form(
            data={
                'name': 'test',
                'tool': 'git',
                'git-path': '/path/to/test.git',
                'git-mirror_path': 'git@localhost:test.git',
                'git-raw_file_url':
                    'https://git.example.com/raw/<revision>/<filename>/',
                'git-username': 'myuser',
                'git-password': 'mypass',
            },
            instance=repository)
        self.assertEqual(form.fields['tool'].initial, 'git')

        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['tool'],
                         scmtools_registry.get_by_id('git'))
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.scmtool_repository_forms['git'],
            ])

        # Make sure none of the other auth forms are unhappy. That would be
        # an indicator that we're doing form processing and validation wrong.
        for auth_form in form.hosting_auth_forms.values():
            self.assertEqual(auth_form.errors, {})

        new_repository = form.save()
        self.assertEqual(repository.pk, new_repository.pk)
        self.assertEqual(repository.extra_data, {})
        self.assertEqual(repository.path, '/path/to/test.git')
        self.assertEqual(repository.mirror_path, 'git@localhost:test.git')
        self.assertEqual(repository.raw_file_url,
                         'https://git.example.com/raw/<revision>/<filename>/')
        self.assertEqual(repository.get_credentials(), {
            'username': 'myuser',
            'password': 'mypass',
        })
        self.assertIsNone(new_repository.local_site)
        self.assertEqual(new_repository.tool,
                         Tool.objects.get(name='Git'))
        self.assertEqual(repository.scmtool_id, 'git')

        self.assertSpyCalledWith(GitTool.check_repository,
                                 path='/path/to/test.git',
                                 username='myuser',
                                 password='mypass',
                                 local_site_name=None)

    def test_plain_repository_with_custom_form_fields(self):
        """Testing RepositoryForm with a plain repository and custom form
        fields
        """
        class CustomAuthForm(StandardSCMToolAuthForm):
            custom_field1 = forms.CharField()
            custom_field2 = forms.BooleanField()

        class CustomRepositoryForm(StandardSCMToolRepositoryForm):
            custom_field3 = forms.CharField()
            custom_field4 = forms.IntegerField()

        old_auth_form = GitTool.auth_form
        old_repository_form = GitTool.repository_form

        GitTool.auth_form = CustomAuthForm
        GitTool.repository_form = CustomRepositoryForm

        try:
            form = self._build_form({
                'name': 'test',
                'tool': 'git',
                'path': '/path/to/test.git',
                'mirror_path': 'git@localhost:test.git',
                'raw_file_url':
                    'https://git.example.com/raw/<revision>/<filename>/',
                'username': 'myuser',
                'password': 'mypass',
                'git-custom_field1': 'str1',
                'git-custom_field2': 'true',
                'git-custom_field3': 'str2',
                'git-custom_field4': '123',
            })
        finally:
            GitTool.auth_form = old_auth_form
            GitTool.repository_form = old_repository_form

        self.assertTrue(form.is_valid())

        repository = form.save()
        self.assertEqual(repository.name, 'test')
        self.assertEqual(repository.tool, Tool.objects.get(name='Git'))
        self.assertEqual(repository.hosting_account, None)
        self.assertEqual(repository.path, '/path/to/test.git')
        self.assertEqual(repository.mirror_path, 'git@localhost:test.git')
        self.assertEqual(repository.raw_file_url,
                         'https://git.example.com/raw/<revision>/<filename>/')
        self.assertEqual(repository.extra_data, {
            'git-custom_field1': 'str1',
            'git-custom_field2': True,
            'git-custom_field3': 'str2',
            'git-custom_field4': 123,
        })
        self.assertEqual(repository.get_credentials(), {
            'username': 'myuser',
            'password': 'mypass',
        })
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.scmtool_repository_forms['git'],
            ])

        # Make sure none of the other auth forms are unhappy. That would be
        # an indicator that we're doing form processing and validation wrong.
        for auth_form in form.hosting_auth_forms.values():
            self.assertEqual(auth_form.errors, {})

        self.assertSpyCalledWith(
            GitTool.check_repository,
            custom_field1='str1',
            custom_field2=True,
            custom_field3='str2',
            custom_field4=123,
            local_site=None,
            local_site_name=None,
            mirror_path='git@localhost:test.git',
            password='mypass',
            path='/path/to/test.git',
            raw_file_url='https://git.example.com/raw/<revision>/<filename>/',
            username='myuser')

    @unittest.skipIf(P4 is None, 'P4 module is not installed')
    def test_plain_repository_with_prefers_mirror_path(self):
        """Testing RepositoryForm with a plain repository and
        SCMTool.prefers_mirror_path=True
        """
        self.assertTrue(PerforceTool.prefers_mirror_path)

        form = self._build_form({
            'name': 'test',
            'tool': 'perforce',
            'path': 'perforce.example.com:1666',
            'mirror_path': 'ssl:perforce.example.com:2666',
            'username': 'myuser',
            'password': 'mypass',
        })

        self.assertTrue(form.is_valid())

        repository = form.save()
        self.assertIsNone(repository.hosting_account)
        self.assertEqual(repository.name, 'test')
        self.assertEqual(repository.tool, Tool.objects.get(name='Perforce'))
        self.assertEqual(repository.scmtool_id, 'perforce')
        self.assertEqual(repository.path, 'perforce.example.com:1666')
        self.assertEqual(repository.mirror_path,
                         'ssl:perforce.example.com:2666')
        self.assertEqual(repository.extra_data, {
            'use_ticket_auth': False,
        })
        self.assertEqual(repository.get_credentials(), {
            'username': 'myuser',
            'password': 'mypass',
        })
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.scmtool_repository_forms['perforce'],
            ])

        # Make sure none of the other auth forms are unhappy. That would be
        # an indicator that we're doing form processing and validation wrong.
        for auth_form in form.hosting_auth_forms.values():
            self.assertEqual(auth_form.errors, {})

        self.assertSpyCalledWith(
            PerforceTool.check_repository,
            local_site=None,
            local_site_name=None,
            mirror_path='ssl:perforce.example.com:2666',
            p4_client=None,
            p4_host=None,
            password='mypass',
            path='perforce.example.com:1666',
            use_ticket_auth=False,
            username='myuser')

    def test_plain_repository_with_missing_fields(self):
        """Testing RepositoryForm with a plain repository with missing fields
        """
        form = self._build_form({
            'name': 'test',
            'tool': 'git',
        })

        self.assertFalse(form.is_valid())
        self.assertIn('path', form.errors)
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.scmtool_repository_forms['git'],
            ])

        # Make sure none of the other auth forms are unhappy. That would be
        # an indicator that we're doing form processing and validation wrong.
        for auth_form in form.hosting_auth_forms.values():
            self.assertEqual(auth_form.errors, {})

    def test_plain_repository_with_unverified_ssl_cert(self):
        """Testing RepositoryForm with a plain repository and unverified
        SSL certificate
        """
        certificate = Certificate(
            fingerprint='4e1243bd22c66e76c2ba9eddc1f91394e57f9f83')

        def _check_repository(cls, *args, **kwargs):
            raise UnverifiedCertificateError(certificate)

        self.spy_on(GitTool.check_repository,
                    owner=GitTool,
                    call_fake=_check_repository)

        form = self._build_form({
            'name': 'test',
            'tool': 'git',
            'path': 'https://git.example.com/',
            'mirror_path': 'git@localhost:test.git',
            'username': 'myuser',
            'password': 'mypass',
            'raw_file_url': 'http://example.com/<revision>',
        })

        self.assertFalse(form.is_valid())
        self.assertIsNone(form.cert)
        self.assertIsNotNone(form.certerror)
        self.assertIsInstance(form.certerror, UnverifiedCertificateError)
        self.assertIs(form.certerror.certificate, certificate)

    def test_plain_repository_with_trust_cert(self):
        """Testing RepositoryForm with a plain repository and trusting SSL
        certificate
        """
        certificate = Certificate(
            fingerprint='4e1243bd22c66e76c2ba9eddc1f91394e57f9f83')
        cert_state = {'valid': False}

        def _check_repository(cls, *args, **kwargs):
            if not cert_state['valid']:
                raise UnverifiedCertificateError(certificate)

        def _accept_certificate(cls, *args, **kwargs):
            cert_state['valid'] = True
            return certificate

        self.spy_on(GitTool.check_repository,
                    owner=GitTool,
                    call_fake=_check_repository)
        self.spy_on(GitTool.accept_certificate,
                    owner=GitTool,
                    call_fake=_accept_certificate)

        form = self._build_form({
            'name': 'test',
            'tool': 'git',
            'path': 'https://git.example.com/',
            'mirror_path': 'git@localhost:test.git',
            'username': 'myuser',
            'password': 'mypass',
            'raw_file_url': 'http://example.com/<revision>',
            'trust_host': 'true',
        })

        self.assertTrue(form.is_valid())
        self.assertIsNone(form.certerror)
        self.assertIs(form.cert, certificate)
        self.assertSpyCalledWith(GitTool.accept_certificate,
                                 path='https://git.example.com/',
                                 username='myuser',
                                 password='mypass',
                                 local_site_name=None,
                                 certificate=certificate)

    @unittest.skipIf(P4 is None, 'P4 module is not installed')
    def test_plain_repository_with_trust_cert_and_prefers_mirror_path(self):
        """Testing RepositoryForm with a plain repository and trusting
        SSL certificate with SCMTool.prefers_mirror_path
        """
        certificate = Certificate(
            fingerprint='4e1243bd22c66e76c2ba9eddc1f91394e57f9f83')
        cert_state = {'valid': False}

        def _check_repository(cls, *args, **kwargs):
            if not cert_state['valid']:
                raise UnverifiedCertificateError(certificate)

        def _accept_certificate(cls, *args, **kwargs):
            cert_state['valid'] = True
            return certificate

        self.spy_on(PerforceTool.check_repository,
                    owner=PerforceTool,
                    call_fake=_check_repository)
        self.spy_on(PerforceTool.accept_certificate,
                    owner=PerforceTool,
                    call_fake=_accept_certificate)

        form = self._build_form({
            'name': 'test',
            'tool': 'perforce',
            'path': 'perforce.example.com:1666',
            'mirror_path': 'ssl:perforce.example.com:2666',
            'username': 'myuser',
            'password': 'mypass',
            'trust_host': 'true',
        })

        self.assertTrue(form.is_valid())
        self.assertIsNone(form.certerror)
        self.assertIs(form.cert, certificate)
        self.assertSpyCalledWith(PerforceTool.accept_certificate,
                                 path='ssl:perforce.example.com:2666',
                                 username='myuser',
                                 password='mypass',
                                 local_site_name=None,
                                 certificate=certificate)

    def test_with_hosting_service_new_account(self):
        """Testing RepositoryForm with a hosting service and new account"""
        form = self._build_form({
            'name': 'test',
            'hosting_type': 'test',
            'test-hosting_account_username': 'testuser',
            'test-hosting_account_password': 'testpass',
            'tool': 'git',
            'test_repo_name': 'testrepo',
        })

        self.assertTrue(form.is_valid())
        self.assertTrue(form.hosting_account_linked)

        repository = form.save()
        self.assertEqual(repository.name, 'test')
        self.assertEqual(repository.scmtool_id, 'git')
        self.assertEqual(repository.hosting_account.username, 'testuser')
        self.assertEqual(repository.hosting_account.service_name, 'test')
        self.assertEqual(repository.hosting_account.local_site, None)
        self.assertEqual(repository.path, 'http://example.com/testrepo/')
        self.assertEqual(repository.mirror_path, '')
        self.assertEqual(
            repository.extra_data,
            {
                'bug_tracker_use_hosting': False,
                'repository_plan': '',
                'test_repo_name': 'testrepo',
            })
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['test']['default'],
            ])

        # Make sure none of the other auth forms are unhappy. That would be
        # an indicator that we're doing form processing and validation wrong.
        for auth_form in form.hosting_auth_forms.values():
            self.assertEqual(auth_form.errors, {})

        self.assertSpyCalledWith(TestService.check_repository,
                                 path='http://example.com/testrepo/',
                                 username=None,
                                 password=None,
                                 local_site_name=None,
                                 plan='default',
                                 scmtool_class=GitTool,
                                 test_repo_name='testrepo',
                                 tool_name='Git')

    def test_with_hosting_service_new_account_auth_error(self):
        """Testing RepositoryForm with a hosting service and new account and
        authorization error
        """
        form = self._build_form({
            'name': 'test',
            'hosting_type': 'test',
            'test-hosting_account_username': 'baduser',
            'test-hosting_account_password': 'testpass',
            'tool': 'git',
            'test_repo_name': 'testrepo',
        })

        self.assertFalse(form.is_valid())
        self.assertFalse(form.hosting_account_linked)
        self.assertIn('hosting_account', form.errors)
        self.assertEqual(form.errors['hosting_account'],
                         ['Unable to link the account: The username is '
                          'very very bad.'])
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['test']['default'],
            ])

        # Make sure none of the other auth forms are unhappy. That would be
        # an indicator that we're doing form processing and validation wrong.
        for auth_form in form.hosting_auth_forms.values():
            self.assertEqual(auth_form.errors, {})

    def test_with_hosting_service_new_account_repo_errors(self):
        """Testing RepositoryForm with a hosting service and new account and
        repository verification errors
        """
        form = self._build_form({
            'name': 'test',
            'hosting_type': 'test',
            'test-hosting_account_username': 'testuser',
            'test-hosting_account_password': 'testpass',
            'test_repo_name': 'bad',
            'tool': 'git',
        })

        self.assertTrue(form.is_valid())
        self.assertTrue(form.hosting_account_linked)

        hosting_account = HostingServiceAccount.objects.get()
        self.assertEqual(form['hosting_account'].value(), hosting_account.pk)

        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['test']['default'],
            ])

        # Make sure none of the other auth forms are unhappy. That would be
        # an indicator that we're doing form processing and validation wrong.
        for auth_form in form.hosting_auth_forms.values():
            self.assertEqual(auth_form.errors, {})

    def test_with_hosting_service_new_account_2fa_code_required(self):
        """Testing RepositoryForm with a hosting service and new account and
        two-factor auth code required
        """
        form = self._build_form({
            'name': 'test',
            'hosting_type': 'test',
            'test-hosting_account_username': '2fa-user',
            'test-hosting_account_password': 'testpass',
            'tool': 'git',
            'test_repo_name': 'testrepo',
        })

        self.assertFalse(form.is_valid())
        self.assertFalse(form.hosting_account_linked)
        self.assertIn('hosting_account', form.errors)
        self.assertEqual(form.errors['hosting_account'],
                         ['Enter your 2FA code.'])
        self.assertTrue(
            form.hosting_service_info['test']['needs_two_factor_auth_code'])
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['test']['default'],
            ])

        # Make sure none of the other auth forms are unhappy. That would be
        # an indicator that we're doing form processing and validation wrong.
        for auth_form in form.hosting_auth_forms.values():
            self.assertEqual(auth_form.errors, {})

    def test_with_hosting_service_new_account_2fa_code_provided(self):
        """Testing RepositoryForm with a hosting service and new account and
        two-factor auth code provided
        """
        form = self._build_form({
            'name': 'test',
            'hosting_type': 'test',
            'test-hosting_account_username': '2fa-user',
            'test-hosting_account_password': 'testpass',
            'test-hosting_account_two_factor_auth_code': '123456',
            'tool': 'git',
            'test_repo_name': 'testrepo',
        })

        self.assertTrue(form.is_valid())
        self.assertTrue(form.hosting_account_linked)
        self.assertFalse(
            form.hosting_service_info['test']['needs_two_factor_auth_code'])
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['test']['default'],
            ])

        # Make sure none of the other auth forms are unhappy. That would be
        # an indicator that we're doing form processing and validation wrong.
        for auth_form in form.hosting_auth_forms.values():
            self.assertEqual(auth_form.errors, {})

    def test_with_hosting_service_new_account_missing_fields(self):
        """Testing RepositoryForm with a hosting service and new account and
        missing fields
        """
        form = self._build_form({
            'name': 'test',
            'hosting_type': 'test',
            'tool': 'git',
            'test_repo_name': 'testrepo',
        })

        self.assertFalse(form.is_valid())
        self.assertFalse(form.hosting_account_linked)
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['test']['default'],
            ])

        self.assertIn('hosting_account_username', form.errors)
        self.assertIn('hosting_account_password', form.errors)

        # Make sure the auth form also contains the errors.
        auth_form = form.hosting_auth_forms.pop('test')
        self.assertIn('hosting_account_username', auth_form.errors)
        self.assertIn('hosting_account_password', auth_form.errors)

        # Make sure none of the other auth forms are unhappy. That would be
        # an indicator that we're doing form processing and validation wrong.
        for auth_form in form.hosting_auth_forms.values():
            self.assertEqual(auth_form.errors, {})

    def test_with_hosting_service_self_hosted_and_new_account(self):
        """Testing RepositoryForm with a self-hosted hosting service and new
        account
        """
        form = self._build_form({
            'name': 'test',
            'hosting_type': 'self_hosted_test',
            'self_hosted_test-hosting_url': 'https://example.com',
            'self_hosted_test-hosting_account_username': 'testuser',
            'self_hosted_test-hosting_account_password': 'testpass',
            'test_repo_name': 'myrepo',
            'tool': 'git',
        })

        self.assertTrue(form.is_valid())
        self.assertTrue(form.hosting_account_linked)
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['self_hosted_test']['default'],
            ])

        repository = form.save()
        self.assertEqual(repository.name, 'test')
        self.assertEqual(repository.scmtool_id, 'git')
        self.assertEqual(repository.hosting_account.hosting_url,
                         'https://example.com')
        self.assertEqual(repository.hosting_account.username, 'testuser')
        self.assertEqual(repository.hosting_account.service_name,
                         'self_hosted_test')
        self.assertEqual(repository.hosting_account.local_site, None)
        self.assertEqual(repository.path, 'https://example.com/myrepo/')
        self.assertEqual(repository.mirror_path, 'git@example.com:myrepo/')
        self.assertEqual(
            repository.extra_data,
            {
                'bug_tracker_use_hosting': False,
                'hosting_url': 'https://example.com',
                'repository_plan': '',
                'test_repo_name': 'myrepo',
            })

        # Make sure none of the other auth forms are unhappy. That would be
        # an indicator that we're doing form processing and validation wrong.
        for auth_form in form.hosting_auth_forms.values():
            self.assertEqual(auth_form.errors, {})

        self.assertSpyCalledWith(SelfHostedTestService.check_repository,
                                 path='https://example.com/myrepo/',
                                 username=None,
                                 password=None,
                                 local_site_name=None,
                                 plan='default',
                                 scmtool_class=GitTool,
                                 test_repo_name='myrepo',
                                 tool_name='Git')

    def test_with_hosting_service_self_hosted_and_blank_url(self):
        """Testing RepositoryForm with a self-hosted hosting service and blank
        URL
        """
        form = self._build_form({
            'name': 'test',
            'hosting_type': 'self_hosted_test',
            'self_hosted_test-hosting_url': '',
            'self_hosted_test-hosting_account_username': 'testuser',
            'self_hosted_test-hosting_account_password': 'testpass',
            'test_repo_name': 'myrepo',
            'tool': 'git',
        })

        self.assertFalse(form.is_valid())
        self.assertFalse(form.hosting_account_linked)
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['self_hosted_test']['default'],
            ])

    def test_with_hosting_service_new_account_localsite(self):
        """Testing RepositoryForm with a hosting service, new account and
        LocalSite
        """
        local_site = LocalSite.objects.create(name='testsite')

        form = self._build_form(
            {
                'name': 'test',
                'hosting_type': 'test',
                'test-hosting_account_username': 'testuser',
                'test-hosting_account_password': 'testpass',
                'tool': 'git',
                'test_repo_name': 'testrepo',
                'local_site': local_site.pk,
            },
            limit_to_local_site=local_site)

        self.assertTrue(form.is_valid())
        self.assertTrue(form.hosting_account_linked)
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['test']['default'],
            ])

        repository = form.save()
        self.assertEqual(repository.name, 'test')
        self.assertEqual(repository.scmtool_id, 'git')
        self.assertEqual(repository.local_site, local_site)
        self.assertEqual(repository.hosting_account.username, 'testuser')
        self.assertEqual(repository.hosting_account.service_name, 'test')
        self.assertEqual(repository.hosting_account.local_site, local_site)
        self.assertEqual(
            repository.extra_data,
            {
                'bug_tracker_use_hosting': False,
                'repository_plan': '',
                'test_repo_name': 'testrepo',
            })

        self.assertSpyCalledWith(TestService.check_repository,
                                 path='http://example.com/testrepo/',
                                 username=None,
                                 password=None,
                                 local_site_name='testsite',
                                 plan='default',
                                 scmtool_class=GitTool,
                                 test_repo_name='testrepo',
                                 tool_name='Git')

    def test_with_hosting_service_existing_account(self):
        """Testing RepositoryForm with a hosting service and existing
        account
        """
        account = HostingServiceAccount.objects.create(username='testuser',
                                                       service_name='test')
        account.data['password'] = 'testpass'
        account.save()

        form = self._build_form({
            'name': 'test',
            'hosting_type': 'test',
            'hosting_account': account.pk,
            'tool': 'git',
            'test_repo_name': 'testrepo',
        })

        self.assertTrue(form.is_valid())
        self.assertFalse(form.hosting_account_linked)
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['test']['default'],
            ])

        repository = form.save()
        self.assertEqual(repository.name, 'test')
        self.assertEqual(repository.scmtool_id, 'git')
        self.assertEqual(repository.hosting_account, account)
        self.assertEqual(
            repository.extra_data,
            {
                'bug_tracker_use_hosting': False,
                'repository_plan': '',
                'test_repo_name': 'testrepo',
            })

        self.assertSpyCalledWith(TestService.check_repository,
                                 path='http://example.com/testrepo/',
                                 username=None,
                                 password=None,
                                 local_site_name=None,
                                 plan='default',
                                 scmtool_class=GitTool,
                                 test_repo_name='testrepo',
                                 tool_name='Git')

    def test_with_hosting_service_existing_account_needs_reauth(self):
        """Testing RepositoryForm with a hosting service and existing
        account needing re-authorization
        """
        # We won't be setting the password, so that is_authorized() will
        # fail.
        account = HostingServiceAccount.objects.create(username='testuser',
                                                       service_name='test')

        form = self._build_form({
            'name': 'test',
            'hosting_type': 'test',
            'hosting_account': account.pk,
            'tool': 'git',
            'test_repo_name': 'testrepo',
        })

        self.assertFalse(form.is_valid())
        self.assertFalse(form.hosting_account_linked)
        self.assertEqual(set(form.errors.keys()),
                         set(['hosting_account_username',
                              'hosting_account_password']))
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['test']['default'],
            ])

    def test_with_hosting_service_existing_account_reauthing(self):
        """Testing RepositoryForm with a hosting service and existing
        account with re-authorizating
        """
        # We won't be setting the password, so that is_authorized() will
        # fail.
        account = HostingServiceAccount.objects.create(username='testuser',
                                                       service_name='test')

        form = self._build_form({
            'name': 'test',
            'hosting_type': 'test',
            'hosting_account': account.pk,
            'test-hosting_account_username': 'testuser2',
            'test-hosting_account_password': 'testpass2',
            'tool': 'git',
            'test_repo_name': 'testrepo',
        })

        self.assertTrue(form.is_valid())
        self.assertTrue(form.hosting_account_linked)
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['test']['default'],
            ])

        account = HostingServiceAccount.objects.get(pk=account.pk)
        self.assertEqual(account.username, 'testuser2')
        self.assertEqual(account.data['password'], 'testpass2')

    def test_with_hosting_service_self_hosted_and_existing_account(self):
        """Testing RepositoryForm with a self-hosted hosting service and
        existing account
        """
        account = HostingServiceAccount.objects.create(
            username='testuser',
            service_name='self_hosted_test',
            hosting_url='https://example.com')
        account.data['password'] = 'testpass'
        account.save()

        form = self._build_form({
            'name': 'test',
            'hosting_type': 'self_hosted_test',
            'self_hosted_test-hosting_url': 'https://example.com',
            'hosting_account': account.pk,
            'tool': 'git',
            'test_repo_name': 'myrepo',
        })

        self.assertTrue(form.is_valid())
        self.assertFalse(form.hosting_account_linked)
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['self_hosted_test']['default'],
            ])

        repository = form.save()
        self.assertEqual(repository.name, 'test')
        self.assertEqual(repository.scmtool_id, 'git')
        self.assertEqual(repository.hosting_account, account)
        self.assertEqual(
            repository.extra_data,
            {
                'bug_tracker_use_hosting': False,
                'hosting_url': 'https://example.com',
                'repository_plan': '',
                'test_repo_name': 'myrepo',
            })

        self.assertSpyCalledWith(SelfHostedTestService.check_repository,
                                 path='https://example.com/myrepo/',
                                 username=None,
                                 password=None,
                                 local_site_name=None,
                                 plan='default',
                                 scmtool_class=GitTool,
                                 test_repo_name='myrepo',
                                 tool_name='Git')

    def test_with_self_hosted_and_invalid_account_service(self):
        """Testing RepositoryForm with a self-hosted hosting service and
        invalid existing account due to mismatched service type
        """
        account = HostingServiceAccount.objects.create(
            username='testuser',
            service_name='self_hosted_test',
            hosting_url='https://example1.com')
        account.data['password'] = 'testpass'
        account.save()

        form = self._build_form({
            'name': 'test',
            'hosting_type': 'test',
            'hosting_account': account.pk,
            'tool': 'git',
            'test_repo_name': 'myrepo',
        })
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['test']['default'],
            ])

        self.assertFalse(form.is_valid())
        self.assertFalse(form.hosting_account_linked)

    def test_with_self_hosted_and_invalid_account_local_site(self):
        """Testing RepositoryForm with a self-hosted hosting service and
        invalid existing account due to mismatched Local Site
        """
        account = HostingServiceAccount.objects.create(
            username='testuser',
            service_name='self_hosted_test',
            hosting_url='https://example1.com',
            local_site=LocalSite.objects.create(name='test-site'))
        account.data['password'] = 'testpass'
        account.save()

        form = self._build_form({
            'name': 'test',
            'hosting_type': 'test',
            'hosting_account': account.pk,
            'tool': 'git',
            'test_repo_name': 'myrepo',
        })

        self.assertFalse(form.is_valid())
        self.assertFalse(form.hosting_account_linked)
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['test']['default'],
            ])

    def test_with_hosting_service_custom_bug_tracker(self):
        """Testing RepositoryForm with a custom bug tracker"""
        account = HostingServiceAccount.objects.create(username='testuser',
                                                       service_name='test')
        account.data['password'] = 'testpass'
        account.save()

        form = self._build_form({
            'name': 'test',
            'hosting_type': 'test',
            'hosting_account': account.pk,
            'tool': 'git',
            'test_repo_name': 'testrepo',
            'bug_tracker_type': 'custom',
            'bug_tracker': 'http://example.com/issue/%s',
        })

        self.assertTrue(form.is_valid())

        repository = form.save()
        self.assertEqual(repository.bug_tracker, 'http://example.com/issue/%s')
        self.assertEqual(repository.scmtool_id, 'git')
        self.assertEqual(
            repository.extra_data,
            {
                'bug_tracker_use_hosting': False,
                'repository_plan': '',
                'test_repo_name': 'testrepo',
            })
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['test']['default'],
            ])

        self.assertSpyCalledWith(TestService.check_repository,
                                 path='http://example.com/testrepo/',
                                 username=None,
                                 password=None,
                                 local_site_name=None,
                                 plan='default',
                                 scmtool_class=GitTool,
                                 test_repo_name='testrepo',
                                 tool_name='Git')

    def test_with_hosting_service_bug_tracker_service(self):
        """Testing RepositoryForm with a bug tracker service"""
        account = HostingServiceAccount.objects.create(username='testuser',
                                                       service_name='test')
        account.data['password'] = 'testpass'
        account.save()

        form = self._build_form({
            'name': 'test',
            'hosting_type': 'test',
            'hosting_account': account.pk,
            'tool': 'git',
            'test_repo_name': 'testrepo',
            'bug_tracker_type': 'test',
            'bug_tracker_hosting_account_username': 'testuser',
            'bug_tracker-test_repo_name': 'testrepo',
        })

        self.assertTrue(form.is_valid())

        repository = form.save()
        self.assertEqual(repository.bug_tracker,
                         'http://example.com/testuser/testrepo/issue/%s')
        self.assertEqual(
            repository.extra_data,
            {
                'bug_tracker_plan': 'default',
                'bug_tracker_type': 'test',
                'bug_tracker_use_hosting': False,
                'bug_tracker-test_repo_name': 'testrepo',
                'bug_tracker-hosting_account_username': 'testuser',
                'repository_plan': '',
                'test_repo_name': 'testrepo',
            })
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['test']['default'],
                form.hosting_bug_tracker_forms['test']['default'],
            ])

        self.assertSpyCalledWith(TestService.check_repository,
                                 path='http://example.com/testrepo/',
                                 local_site_name=None,
                                 plan='default',
                                 scmtool_class=GitTool,
                                 test_repo_name='testrepo',
                                 tool_name='Git')

    def test_with_hosting_service_self_hosted_bug_tracker_service(self):
        """Testing RepositoryForm with a self-hosted bug tracker service"""
        account = HostingServiceAccount.objects.create(
            username='testuser',
            service_name='self_hosted_test',
            hosting_url='https://example.com')
        account.data['password'] = 'testpass'
        account.save()

        form = self._build_form({
            'name': 'test',
            'hosting_type': 'self_hosted_test',
            'hosting_url': 'https://example.com',
            'hosting_account': account.pk,
            'tool': 'git',
            'test_repo_name': 'testrepo',
            'bug_tracker_type': 'self_hosted_test',
            'bug_tracker_hosting_url': 'https://example.com',
            'bug_tracker-test_repo_name': 'testrepo',
        })

        self.assertTrue(form.is_valid())

        repository = form.save()
        self.assertEqual(repository.bug_tracker,
                         'https://example.com/testrepo/issue/%s')
        self.assertEqual(
            repository.extra_data,
            {
                'bug_tracker_hosting_url': 'https://example.com',
                'bug_tracker_plan': 'default',
                'bug_tracker_type': 'self_hosted_test',
                'bug_tracker_use_hosting': False,
                'bug_tracker-test_repo_name': 'testrepo',
                'hosting_url': 'https://example.com',
                'repository_plan': '',
                'test_repo_name': 'testrepo',
            })
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['self_hosted_test']['default'],
                form.hosting_bug_tracker_forms['self_hosted_test']['default'],
            ])

        self.assertSpyCalledWith(SelfHostedTestService.check_repository,
                                 path='https://example.com/testrepo/',
                                 username=None,
                                 password=None,
                                 local_site_name=None,
                                 plan='default',
                                 scmtool_class=GitTool,
                                 test_repo_name='testrepo',
                                 tool_name='Git')

    def test_with_hosting_service_with_hosting_bug_tracker(self):
        """Testing RepositoryForm with hosting service's bug tracker"""
        account = HostingServiceAccount.objects.create(username='testuser',
                                                       service_name='github')
        account.data['authorization'] = {
            'token': 'abc123',
        }
        account.save()

        form = self._build_form({
            'name': 'test',
            'hosting_type': 'github',
            'hosting_account': account.pk,
            'repository_plan': 'public',
            'tool': 'git',
            'github_public_repo_name': 'testrepo',
            'bug_tracker_use_hosting': True,
            'bug_tracker_type': 'github',
            'bug_tracker_plan': 'public',
        })

        self.assertTrue(form.is_valid())
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['github']['public'],
            ])

        repository = form.save()
        self.assertEqual(repository.bug_tracker,
                         'http://github.com/testuser/testrepo/issues#issue/%s')
        self.assertEqual(
            repository.extra_data,
            {
                'bug_tracker_use_hosting': True,
                'github_public_repo_name': 'testrepo',
                'repository_plan': 'public',
            })

        self.assertSpyCalledWith(
            GitHub.check_repository,
            path='git://github.com/testuser/testrepo.git',
            username=None,
            password=None,
            local_site_name=None,
            plan='public',
            scmtool_class=GitTool,
            github_public_repo_name='testrepo',
            tool_name='Git')

    def test_with_hosting_service_with_hosting_bug_tracker_and_self_hosted(
            self):
        """Testing RepositoryForm with self-hosted hosting service's bug
        tracker
        """
        account = HostingServiceAccount.objects.create(
            username='testuser',
            service_name='self_hosted_test',
            hosting_url='https://example.com')
        account.data['password'] = 'testpass'
        account.save()

        account.data['authorization'] = {
            'token': '1234',
        }
        account.save()

        form = self._build_form({
            'name': 'test',
            'hosting_type': 'self_hosted_test',
            'hosting_url': 'https://example.com',
            'hosting_account': account.pk,
            'tool': 'git',
            'test_repo_name': 'testrepo',
            'bug_tracker_use_hosting': True,
            'bug_tracker_type': 'github',
        })

        self.assertTrue(form.is_valid())
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['self_hosted_test']['default'],
            ])

        repository = form.save()
        self.assertEqual(repository.bug_tracker,
                         'https://example.com/testrepo/issue/%s')
        self.assertEqual(
            repository.extra_data,
            {
                'bug_tracker_use_hosting': True,
                'hosting_url': 'https://example.com',
                'repository_plan': '',
                'test_repo_name': 'testrepo',
            })

        self.assertSpyCalledWith(SelfHostedTestService.check_repository,
                                 path='https://example.com/testrepo/',
                                 username=None,
                                 password=None,
                                 local_site_name=None,
                                 plan='default',
                                 scmtool_class=GitTool,
                                 test_repo_name='testrepo',
                                 tool_name='Git')

    def test_with_hosting_service_no_bug_tracker(self):
        """Testing RepositoryForm with no bug tracker"""
        account = HostingServiceAccount.objects.create(username='testuser',
                                                       service_name='test')
        account.data['password'] = 'testpass'
        account.save()

        form = self._build_form({
            'name': 'test',
            'hosting_type': 'test',
            'hosting_account': account.pk,
            'tool': 'git',
            'test_repo_name': 'testrepo',
        })

        self.assertTrue(form.is_valid())
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['test']['default'],
            ])

        repository = form.save()
        self.assertEqual(repository.bug_tracker, '')
        self.assertEqual(
            repository.extra_data,
            {
                'bug_tracker_use_hosting': False,
                'repository_plan': '',
                'test_repo_name': 'testrepo',
            })

        self.assertSpyCalledWith(TestService.check_repository,
                                 path='http://example.com/testrepo/',
                                 username=None,
                                 password=None,
                                 local_site_name=None,
                                 plan='default',
                                 scmtool_class=GitTool,
                                 test_repo_name='testrepo',
                                 tool_name='Git')

    def test_with_hosting_service_with_existing_custom_bug_tracker(self):
        """Testing RepositoryForm with existing custom bug tracker"""
        repository = self.create_repository(
            name='test',
            bug_tracker='http://example.com/issue/%s')

        form = RepositoryForm(instance=repository)
        self.assertFalse(form._get_field_data('bug_tracker_use_hosting'))
        self.assertEqual(form._get_field_data('bug_tracker_type'), 'custom')
        self.assertEqual(form.initial['bug_tracker'],
                         'http://example.com/issue/%s')
        self.assertEqual(list(form.iter_subforms(bound_only=True)), [])

    def test_with_hosting_service_with_existing_bug_tracker_service(self):
        """Testing RepositoryForm with existing bug tracker service"""
        repository = self.create_repository(
            name='test',
            extra_data={
                'bug_tracker_type': 'test',
                'bug_tracker-hosting_account_username': 'testuser',
                'bug_tracker-test_repo_name': 'testrepo',
            })

        form = RepositoryForm(instance=repository)
        self.assertFalse(form._get_field_data('bug_tracker_use_hosting'))
        self.assertEqual(form._get_field_data('bug_tracker_type'), 'test')
        self.assertEqual(
            form._get_field_data('bug_tracker_hosting_account_username'),
            'testuser')

        self.assertIn('test', form.hosting_bug_tracker_forms)
        self.assertIn('default', form.hosting_bug_tracker_forms['test'])
        bitbucket_form = form.hosting_bug_tracker_forms['test']['default']
        self.assertEqual(
            bitbucket_form.fields['test_repo_name'].initial,
            'testrepo')

    def test_with_hosting_service_with_existing_bug_tracker_using_hosting(
            self):
        """Testing RepositoryForm with existing bug tracker using hosting
        service
        """
        account = HostingServiceAccount.objects.create(username='testuser',
                                                       service_name='test')
        repository = self.create_repository(
            name='test',
            hosting_account=account,
            extra_data={
                'bug_tracker_use_hosting': True,
                'test_repo_name': 'testrepo',
            })

        form = RepositoryForm(instance=repository)
        self.assertTrue(form._get_field_data('bug_tracker_use_hosting'))

    def test_bound_forms_with_post_with_repository_service(self):
        """Testing RepositoryForm binds hosting service forms only if matching
        posted repository hosting_service using default plan
        """
        form = self._build_form({
            'name': 'test',
            'hosting_type': 'test',
        })

        # Make sure only the relevant forms are bound.
        self.assertEqual(
            list(form.iter_subforms(bound_only=True,
                                    with_auth_forms=True)),
            [
                form.hosting_repository_forms['test']['default'],
            ])

    def test_bound_forms_with_post_with_bug_tracker_service(self):
        """Testing RepositoryForm binds hosting service forms only if matching
        posted bug tracker hosting_service using default plan
        """
        form = self._build_form({
            'name': 'test',
            'bug_tracker_type': 'test',
        })

        # Make sure only the relevant forms are bound.
        self.assertEqual(
            list(form.iter_subforms(bound_only=True,
                                    with_auth_forms=True)),
            [
                form.hosting_bug_tracker_forms['test']['default'],
            ])

    def test_bound_forms_with_post_with_repo_service_and_plan(self):
        """Testing RepositoryForm binds hosting service forms only if matching
        posted repository hosting_service with specific plans
        """
        form = self._build_form({
            'name': 'test',
            'hosting_type': 'github',
            'repository_plan': 'public',
        })

        # Make sure only the relevant forms are bound.
        self.assertEqual(
            list(form.iter_subforms(bound_only=True,
                                    with_auth_forms=True)),
            [
                form.hosting_repository_forms['github']['public'],
            ])

    def test_bound_forms_with_post_with_bug_tracker_service_and_plan(self):
        """Testing RepositoryForm binds hosting service forms only if matching
        posted bug tracker hosting_service with specific plans
        """
        form = self._build_form({
            'name': 'test',
            'bug_tracker_type': 'github',
            'bug_tracker_plan': 'public',
        })

        # Make sure only the relevant forms are bound.
        self.assertEqual(
            list(form.iter_subforms(bound_only=True,
                                    with_auth_forms=True)),
            [
                form.hosting_bug_tracker_forms['github']['public'],
            ])

    def test_with_set_access_list(self):
        """Testing RepositoryForm with setting users access list"""
        user1 = User.objects.create(username='user1')
        user2 = User.objects.create(username='user2')
        User.objects.create(username='user3')

        group1 = self.create_review_group(name='group1', invite_only=True)
        group2 = self.create_review_group(name='group2', invite_only=True)
        self.create_review_group(name='group3', invite_only=True)

        form = self._build_form({
            'name': 'test',
            'tool': 'git',
            'path': '/path/to/test.git',
            'public': False,
            'users': [user1.pk, user2.pk],
            'review_groups': [group1.pk, group2.pk],
        })

        form.is_valid()
        self.assertTrue(form.is_valid())

        repository = form.save()
        self.assertFalse(repository.public)
        self.assertEqual(repository.extra_data, {})
        self.assertEqual(list(repository.users.all()), [user1, user2])
        self.assertEqual(list(repository.review_groups.all()),
                         [group1, group2])
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.scmtool_repository_forms['git'],
            ])

    def test_with_set_public_and_prev_access_list(self):
        """Testing RepositoryForm with setting public=True when an access list
        is set
        """
        user = self.create_user()
        review_group = self.create_review_group(invite_only=True)

        repository = self.create_repository(tool_name='Test',
                                            public=False)
        repository.users.add(user)
        repository.review_groups.add(review_group)

        form = self._build_form(
            {
                'name': 'test',
                'tool': 'git',
                'path': '/path/to/test.git',
                'public': True,
                'users': [user.pk],
                'review_groups': [review_group.pk],
            },
            instance=repository)
        self.assertEqual(form['users'].value(), [user.pk])
        self.assertEqual(form['review_groups'].value(), [review_group.pk])

        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['users'], [])
        self.assertEqual(form.cleaned_data['review_groups'], [])

        repository = form.save()
        self.assertEqual(repository.users.count(), 0)
        self.assertEqual(repository.review_groups.count(), 0)

    def test_public_checkbox_with_login_required(self):
        """Testing RepositoryForm public checkbox with site-wide login required
        """
        with self.siteconfig_settings({'auth_require_sitewide_login': True}):
            form = RepositoryForm()
            field = form.fields['public']

            self.assertEqual(field.label, 'Accessible to all logged-in users')
            self.assertEqual(
                field.help_text,
                'Review requests and files on this repository will be '
                'visible to any logged-in users. Uncheck this box to '
                'grant access only to specific users and/or to users '
                'who are members of specific invite-only review groups.')

    def test_public_checkbox_with_login_not_required(self):
        """Testing RepositoryForm public checkbox with site-wide login not
        required
        """
        with self.siteconfig_settings({'auth_require_sitewide_login': False}):
            form = RepositoryForm()
            field = form.fields['public']

            self.assertEqual(field.label, 'Accessible to everyone')
            self.assertEqual(
                field.help_text,
                'Review requests and files on this repository will be '
                'visible to any anonymous or logged-in users. Uncheck '
                'this box to grant access only to specific users and/or '
                'to users who are members of specific invite-only '
                'review groups.')

    def test_public_checkbox_with_limit_local_site_not_public(self):
        """Testing RepositoryForm public checkbox with form limited to
        LocalSite and site not public
        """
        with self.siteconfig_settings({'auth_require_sitewide_login': True}):
            local_site = LocalSite.objects.create(name='test-site')
            form = RepositoryForm(limit_to_local_site=local_site)
            field = form.fields['public']

            self.assertEqual(field.label,
                             'Accessible to all users on test-site')
            self.assertEqual(
                field.help_text,
                'Review requests and files on this repository will be '
                'visible to anyone on test-site. Uncheck this box to grant '
                'access only to specific users and/or to users who are '
                'members of specific invite-only review groups.')

    def test_public_checkbox_with_limit_local_site_public(self):
        """Testing RepositoryForm public checkbox with form limited to
        LocalSite and site is public
        """
        with self.siteconfig_settings({'auth_require_sitewide_login': True}):
            local_site = LocalSite.objects.create(name='test-site',
                                                  public=True)
            form = RepositoryForm(limit_to_local_site=local_site)
            field = form.fields['public']

            self.assertEqual(field.label, 'Accessible to all logged-in users')
            self.assertEqual(
                field.help_text,
                'Review requests and files on this repository will be '
                'visible to any logged-in users. Uncheck this box to '
                'grant access only to specific users and/or to users '
                'who are members of specific invite-only review groups.')

    def test_public_checkbox_with_instance_local_site_not_public(self):
        """Testing RepositoryForm public checkbox with LocalSite-owned
        repository and site not public
        """
        with self.siteconfig_settings({'auth_require_sitewide_login': True}):
            local_site = LocalSite.objects.create(name='test-site')
            repository = self.create_repository(tool_name='Test',
                                                local_site=local_site)
            form = RepositoryForm(instance=repository)
            field = form.fields['public']

            self.assertEqual(field.label,
                             'Accessible to all users on test-site')
            self.assertEqual(
                field.help_text,
                'Review requests and files on this repository will be '
                'visible to anyone on test-site. Uncheck this box to grant '
                'access only to specific users and/or to users who are '
                'members of specific invite-only review groups.')

    def test_public_checkbox_with_instance_local_site_public(self):
        """Testing RepositoryForm public checkbox with LocalSite-owned
        repository and site is public
        """
        with self.siteconfig_settings({'auth_require_sitewide_login': True}):
            local_site = LocalSite.objects.create(name='test-site',
                                                  public=True)
            repository = self.create_repository(tool_name='Test',
                                                local_site=local_site)
            form = RepositoryForm(instance=repository)
            field = form.fields['public']

            self.assertEqual(field.label, 'Accessible to all logged-in users')
            self.assertEqual(
                field.help_text,
                'Review requests and files on this repository will be '
                'visible to any logged-in users. Uncheck this box to '
                'grant access only to specific users and/or to users '
                'who are members of specific invite-only review groups.')

    def test_extra_data_with_new_repo(self):
        """Testing RepositoryForm preserves extra_data by default on new
        repositories
        """
        account = HostingServiceAccount.objects.create(username='testuser',
                                                       service_name='github')
        account.data['authorization'] = {
            'token': 'abc123',
        }
        account.save()

        form = self._build_form({
            'name': 'test',
            'hosting_type': 'github',
            'hosting_account': account.pk,
            'repository_plan': 'public',
            'tool': 'git',
            'github_public_repo_name': 'testrepo',
            'bug_tracker_use_hosting': True,
            'bug_tracker_type': 'github',
            'bug_tracker_plan': 'public',
            'extra_data': {
                'test-key': 'test-value',
                'another-key': 123,
            },
        })

        self.assertTrue(form.is_valid())

        repository = form.save()
        self.assertEqual(repository.name, 'test')
        self.assertEqual(repository.tool, Tool.objects.get(name='Git'))
        self.assertEqual(repository.scmtool_id, 'git')
        self.assertEqual(repository.hosting_account, account)
        self.assertEqual(repository.extra_data, {
            'another-key': 123,
            'bug_tracker_use_hosting': True,
            'github_public_repo_name': 'testrepo',
            'repository_plan': 'public',
            'test-key': 'test-value',
        })
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['github']['public'],
            ])

    def test_extra_data_with_existing_repo(self):
        """Testing RepositoryForm preserves extra_data by default on existing
        repositories
        """
        account = HostingServiceAccount.objects.create(username='testuser',
                                                       service_name='github')
        account.data['authorization'] = {
            'token': 'abc123',
        }
        account.save()

        repository = self.create_repository(name='Test Repository',
                                            path='/path/',
                                            mirror_path='/mirror_path/',
                                            raw_file_url='/raw_file/',
                                            encoding='utf-128',
                                            tool_name='Git',
                                            hosting_account=account)
        repository.extra_data.update({
            'github_public_repo_name': 'testrepo',
            'bug_tracker_hosting_url': 'http://example.com/',
            'bug_tracker_type': 'github',
            'bug_tracker-hosting_account_username': 'test-user',
            'bug_tracker-github_repo_name': 'test-repo',
            'bug_tracker_plan': 'private',
            'repository_plan': 'public',
            'test-key': 'test-value',
            'another-key': 123,
        })

        form = self._build_form(
            {
                'name': 'test',
                'hosting_type': 'github',
                'hosting_account': account.pk,
                'repository_plan': 'private',
                'tool': 'git',
                'github_private_repo_name': 'testrepo',
                'bug_tracker_use_hosting': True,
                'bug_tracker_type': 'github',
                'bug_tracker_plan': 'private',
                'extra_data': repository.extra_data,
            },
            instance=repository)

        self.assertTrue(form.is_valid())

        repository = form.save()
        self.assertEqual(repository.name, 'test')
        self.assertEqual(repository.tool, Tool.objects.get(name='Git'))
        self.assertEqual(repository.scmtool_id, 'git')
        self.assertEqual(repository.hosting_account, account)
        self.assertEqual(repository.extra_data, {
            'another-key': 123,
            'bug_tracker_use_hosting': True,
            'github_private_repo_name': 'testrepo',
            'repository_plan': 'private',
            'test-key': 'test-value',
        })
        self.assertEqual(
            list(form.iter_subforms(bound_only=True)),
            [
                form.hosting_repository_forms['github']['private'],
            ])

    def test_extra_data_with_invalid_type(self):
        """Testing RepositoryForm validates extra_data as dictionary"""
        form = self._build_form({
            'name': 'test',
            'tool': 'git',
            'path': '/path/to/repo.git',
            'extra_data': [1, 2, 3],
        })

        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors['extra_data'],
                         ['This must be a JSON object/dictionary.'])

    def test_skips_hosting_service_with_visible_services(self):
        """Testing RepositoryForm shows only visible HostingServices"""
        form = RepositoryForm()

        hosting_types = {
            key
            for key, value in form.fields['hosting_type'].choices
        }

        self.assertIn('test', hosting_types)
        self.assertNotIn('hidden-test', hosting_types)

        self.assertIn('test', form.hosting_service_info)
        self.assertNotIn('hidden-test', form.hosting_service_info)

    def test_skips_hosting_service_with_visible_services_and_instance(self):
        """Testing RepositoryForm shows hidden HostingService if set by
        instance
        """
        hosting_account = HostingServiceAccount.objects.create(
            username='test-user',
            service_name='hidden-test')
        repository = self.create_repository(hosting_account=hosting_account,
                                            tool_name='Git')
        form = RepositoryForm(instance=repository)

        hosting_types = {
            key
            for key, value in form.fields['hosting_type'].choices
        }

        self.assertIn('test', hosting_types)
        self.assertIn('hidden-test', hosting_types)

        self.assertIn('test', form.hosting_service_info)
        self.assertIn('hidden-test', form.hosting_service_info)

    def test_hosting_service_info_with_visible_scms(self):
        """Testing RepositoryForm.hosting_service_info contains visible
        SCMTools
        """
        form = RepositoryForm()

        self.assertEqual(form.hosting_service_info['test']['scmtools'],
                         ['git', 'test'])

    def test_hosting_service_info_with_visible_scms_and_instance(self):
        """Testing RepositoryForm.hosting_service_info contains both
        visible SCMTools and instance's service
        """
        hosting_account = HostingServiceAccount.objects.create(
            username='test-user',
            service_name='test')
        repository = self.create_repository(hosting_account=hosting_account,
                                            tool_name='Perforce')
        form = RepositoryForm(instance=repository)

        self.assertEqual(form.hosting_service_info['test']['scmtools'],
                         ['git', 'perforce', 'test'])

    def _build_form(self, data=None, check_repository=False, **kwargs):
        """Build the repository form with some standard data.

        This will pre-fill any supplied data will defaults based on the form's
        initial data, and also supports disabling repository checks.

        Args:
            data (dict, optional):
                Posted data to provide to the form. If supplied, it will also
                consist of defaults from the form.

            check_repository (bool, optional):
                Whether to check the validity of repositories.

            **kwargs (dict, optional):
                Additional keyword arguments to pass to the form.

        Returns:
            reviewboard.scmtools.forms.RepositoryForm:
            The form instance.
        """
        if data is not None:
            post_data = QueryDict(mutable=True)
            post_data.update(dict({
                name: field.initial
                for name, field in RepositoryForm.base_fields.items()
            }, **data))
            post_data._mutable = False
        else:
            post_data = None

        form = RepositoryForm(post_data, **kwargs)

        if post_data is not None and not check_repository:
            hosting_type = post_data['hosting_type']
            tool_id = post_data['tool']

            if hosting_type != 'custom':
                hosting_service = \
                    hosting_service_registry.get_hosting_service(hosting_type)

                if not hasattr(hosting_service.check_repository, 'spy'):
                    self.spy_on(hosting_service.check_repository,
                                owner=hosting_service,
                                call_original=False)
            elif tool_id:
                scmtool_cls = scmtools_registry.get_by_id(tool_id)

                if not hasattr(scmtool_cls.check_repository, 'spy'):
                    self.spy_on(scmtool_cls.check_repository,
                                owner=scmtool_cls,
                                call_original=False)

        return form



"""Tests for reviewboard.diffviewer.views.DiffFragmentView."""

from __future__ import annotations

from django.contrib.auth.models import User
from django.core.cache import cache

from reviewboard.attachments.models import FileAttachment
from reviewboard.attachments.tests.base import BaseFileAttachmentTestCase
from reviewboard.site.urlresolvers import local_site_reverse
from reviewboard.testing import TestCase


class ReviewsDiffFragmentViewTests(TestCase):
    """Tests for reviewboard.diffviewer.views.DiffFragmentView."""

    fixtures = ['test_scmtools', 'test_users']

    def test_base_filediff_not_in_diffset(self):
        """Testing ReviewsDiffFragmentView.get with ?base-filediff-id= as a
        FileDiff outside the current diffset
        """
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(repository=repository,
                                                    create_with_history=True)
        review_request.target_people.add(review_request.submitter)

        diffset = self.create_diffset(review_request, draft=True)
        commit = self.create_diffcommit(diffset=diffset)
        diffset.finalize_commit_series(
            cumulative_diff=self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
            validation_info=None,
            validate=False,
            save=True)

        review_request.publish(user=review_request.submitter)

        filediff = commit.files.get()

        other_diffset = self.create_diffset(repository=repository)
        other_filediff = self.create_filediff(diffset=other_diffset)

        rsp = self.client.get(
            local_site_reverse(
                'view-diff-fragment',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset.revision,
                    'filediff_id': filediff.pk,
                }),
            data={'base-filediff-id': other_filediff.pk})

        self.assertEqual(rsp.status_code, 404)

    def test_base_filediff_and_interfilediff(self):
        """Testing ReviewsDiffFragmentView.get with interfilediff and
        ?base-filediff-id
        """
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(repository=repository,
                                                    create_with_history=True)
        review_request.target_people.add(review_request.submitter)

        diffset = self.create_diffset(review_request, draft=True)
        diffset_commits = [
            self.create_diffcommit(diffset=diffset, commit_id='r1',
                                   parent_id='r0'),
            self.create_diffcommit(diffset=diffset, commit_id='r2',
                                   parent_id='r1'),
        ]

        filediff = diffset_commits[1].files.get()
        base_filediff = diffset_commits[0].files.get()

        diffset.finalize_commit_series(
            cumulative_diff=self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
            validation_info=None,
            validate=False,
            save=True)
        review_request.publish(user=review_request.submitter)

        interdiffset = self.create_diffset(review_request, draft=True)
        interdiffset_commit = self.create_diffcommit(
            diffset=interdiffset, commit_id='r1', parent_id='r0')

        interdiffset.finalize_commit_series(
            cumulative_diff=self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
            validation_info=None,
            validate=False,
            save=True)
        review_request.publish(user=review_request.submitter)

        interfilediff = interdiffset_commit.files.get()

        rsp = self.client.get(
            local_site_reverse(
                'view-diff-fragment',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset.revision,
                    'interdiff_revision': interdiffset.revision,
                    'filediff_id': filediff.pk,
                    'interfilediff_id': interfilediff.pk,
                }),
            data={'base-filediff-id': base_filediff.pk})

        self.assertEqual(rsp.status_code, 500)
        self.assertIn(
            b'Cannot generate an interdiff when base FileDiff ID is '
            b'specified.',
            rsp.content)

    def test_base_filediff_not_ancestor(self):
        """Testing ReviewsDiffFragmentView.get with ?base-filediff-id set to a
        FileDiff that is not an ancestor FileDiff
        """
        review_request = self.create_review_request(create_repository=True,
                                                    create_with_history=True)
        review_request.target_people.add(review_request.submitter)
        diffset = self.create_diffset(review_request, draft=True)
        commits = [
            self.create_diffcommit(diffset=diffset, commit_id='r1',
                                   parent_id='r0'),
            self.create_diffcommit(diffset=diffset, commit_id='r2',
                                   parent_id='r1'),
        ]

        base_filediff = commits[0].files.get()
        filediff = commits[1].files.get()

        diffset.finalize_commit_series(
            cumulative_diff=self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
            validation_info=None,
            validate=False,
            save=True)
        review_request.publish(user=review_request.submitter)

        rsp = self.client.get(
            local_site_reverse(
                'view-diff-fragment',
                kwargs={
                    'revision': diffset.revision,
                    'review_request_id': review_request.display_id,
                    'filediff_id': filediff.pk,
                }),
            data={'base-filediff-id': base_filediff.pk})

        self.assertEqual(rsp.status_code, 500)
        self.assertIn(
            b'The requested FileDiff (ID %d) is not a valid base FileDiff '
            b'for FileDiff %d.'
            % (base_filediff.pk, filediff.pk),
            rsp.content)


class ReviewsFileAttachmentDiffFragmentViewTests(BaseFileAttachmentTestCase):
    """Tests for inline diff file attachments in the diff viewer.

    Version Added:
        7.0.3:
        This was split off from :py:mod:`reviewboard.attachments.tests`.
    """

    def setUp(self):
        """Set up this test case."""
        super().setUp()

        # The diff viewer's caching breaks the result of these tests,
        # so be sure we clear before each one.
        cache.clear()

    def test_added_file(self):
        """Testing inline diff file attachments with newly added files"""
        # Set up the initial state.
        user = User.objects.get(username='doc')
        review_request = self.create_review_request(submitter=user,
                                                    target_people=[user])
        filediff = self.make_filediff(
            is_new=True,
            diffset_history=review_request.diffset_history)

        # Create a diff file attachment to be displayed inline.
        diff_file_attachment = FileAttachment.objects.create_from_filediff(
            filediff,
            orig_filename='my-file',
            file=self.make_uploaded_file(),
            mimetype='image/png')
        review_request.file_attachments.add(diff_file_attachment)
        review_request.save()
        review_request.publish(user)

        # Load the diff viewer.
        self.client.login(username='doc', password='doc')
        response = self.client.get('/r/%d/diff/1/fragment/%s/'
                                   % (review_request.pk, filediff.pk))
        self.assertEqual(response.status_code, 200)

        # The file attachment should appear as the right-hand side
        # file attachment in the diff viewer.
        self.assertEqual(response.context['orig_diff_file_attachment'], None)
        self.assertEqual(response.context['modified_diff_file_attachment'],
                         diff_file_attachment)

    def test_modified_file(self):
        """Testing inline diff file attachments with modified files"""
        # Set up the initial state.
        user = User.objects.get(username='doc')
        review_request = self.create_review_request(submitter=user)
        filediff = self.make_filediff(
            is_new=False,
            diffset_history=review_request.diffset_history)
        self.assertFalse(filediff.is_new)

        # Create diff file attachments to be displayed inline.
        uploaded_file = self.make_uploaded_file()

        orig_attachment = FileAttachment.objects.create_from_filediff(
            filediff,
            orig_filename='my-file',
            file=uploaded_file,
            mimetype='image/png',
            from_modified=False)
        modified_attachment = FileAttachment.objects.create_from_filediff(
            filediff,
            orig_filename='my-file',
            file=uploaded_file,
            mimetype='image/png')
        review_request.file_attachments.add(orig_attachment)
        review_request.file_attachments.add(modified_attachment)
        review_request.publish(user)

        # Load the diff viewer.
        self.client.login(username='doc', password='doc')
        response = self.client.get('/r/%d/diff/1/fragment/%s/'
                                   % (review_request.pk, filediff.pk))
        self.assertEqual(response.status_code, 200)

        # The file attachment should appear as the right-hand side
        # file attachment in the diff viewer.
        self.assertEqual(response.context['orig_diff_file_attachment'],
                         orig_attachment)
        self.assertEqual(response.context['modified_diff_file_attachment'],
                         modified_attachment)

    def test_etag_match(self) -> None:
        """Testing ETag generation for binary file fragment with comments"""
        user = User.objects.get(username='doc')
        review_request = self.create_review_request(submitter=user)
        filediff = self.make_filediff(
            is_new=False,
            diffset_history=review_request.diffset_history)

        # Create file attachments for the binary file.
        uploaded_file = self.make_uploaded_file()
        orig_attachment = FileAttachment.objects.create_from_filediff(
            filediff,
            orig_filename='binary-file.png',
            file=uploaded_file,
            mimetype='image/png',
            from_modified=False)
        modified_attachment = FileAttachment.objects.create_from_filediff(
            filediff,
            orig_filename='binary-file.png',
            file=uploaded_file,
            mimetype='image/png')
        review_request.file_attachments.add(orig_attachment)
        review_request.file_attachments.add(modified_attachment)
        review_request.publish(user)

        # Create a review with file attachment comment.
        review = self.create_review(
            review_request=review_request,
            user=user,
            public=True)

        self.create_file_attachment_comment(
            review=review,
            file_attachment=modified_attachment,
            diff_against_file_attachment=orig_attachment)

        # Request the fragment.
        url = f'/r/{review_request.pk}/diff/1/fragment/{filediff.pk}/'

        self.client.login(username='doc', password='doc')
        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)

        etag = response.headers['ETag']

        # Fetch the fragment again with the same ETag. We should get a 304.
        response = self.client.get(url, headers={'If-None-Match': etag})
        self.assertEqual(response.status_code, 304)

    def test_etag_new_comments(self) -> None:
        """Testing ETag updates for binary file fragment with newly-added
        comments
        """
        user = User.objects.get(username='doc')
        review_request = self.create_review_request(submitter=user)
        filediff = self.make_filediff(
            is_new=False,
            diffset_history=review_request.diffset_history)

        # Create file attachments for the binary file.
        uploaded_file = self.make_uploaded_file()
        orig_attachment = FileAttachment.objects.create_from_filediff(
            filediff,
            orig_filename='binary-file.png',
            file=uploaded_file,
            mimetype='image/png',
            from_modified=False)
        modified_attachment = FileAttachment.objects.create_from_filediff(
            filediff,
            orig_filename='binary-file.png',
            file=uploaded_file,
            mimetype='image/png')
        review_request.file_attachments.add(orig_attachment)
        review_request.file_attachments.add(modified_attachment)
        review_request.publish(user)

        # Create a review with file attachment comment.
        review = self.create_review(
            review_request=review_request,
            user=user)

        self.create_file_attachment_comment(
            review=review,
            file_attachment=modified_attachment,
            diff_against_file_attachment=orig_attachment)

        # Request the fragment.
        url = f'/r/{review_request.pk}/diff/1/fragment/{filediff.pk}/'

        self.client.login(username='doc', password='doc')
        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)

        etag = response.headers['ETag']

        # Create a new comment.
        self.create_file_attachment_comment(
            review=review,
            file_attachment=modified_attachment,
            diff_against_file_attachment=orig_attachment)

        # Fetch the fragment again with the same ETag. We should now get a 200
        # instead of a 304.
        response = self.client.get(url, headers={'If-None-Match': etag})
        self.assertEqual(response.status_code, 200)



"""Reduce max_length of HostingServiceAccount.hosting_url to 255 chars.

This was reduced from 256 characters.

Version Added:
    1.7.8
"""

from django_evolution.mutations import ChangeField


MUTATIONS = [
    ChangeField('HostingServiceAccount', 'hosting_url', initial=None,
                max_length=255)
]



"""Unit tests for reviewboard.extensions.hooks.AvatarServiceHook."""

from djblets.avatars.tests import DummyAvatarService

from reviewboard.avatars import avatar_services
from reviewboard.extensions.hooks import AvatarServiceHook
from reviewboard.extensions.tests.testcases import BaseExtensionHookTestCase


class AvatarServiceHookTests(BaseExtensionHookTestCase):
    """Test for reviewboard.extensions.hooks.AvatarServiceHook."""

    @classmethod
    def setUpClass(cls):
        super(AvatarServiceHookTests, cls).setUpClass()
        avatar_services.reset()

    def tearDown(self):
        super(AvatarServiceHookTests, self).tearDown()
        avatar_services.reset()

    def test_register(self):
        """Testing AvatarServiceHook registers services"""
        self.assertNotIn(DummyAvatarService, avatar_services)
        AvatarServiceHook(self.extension, DummyAvatarService,
                          start_enabled=True)
        self.assertIn(DummyAvatarService, avatar_services)

        avatar_services.enable_service(DummyAvatarService, save=False)
        self.assertTrue(avatar_services.is_enabled(DummyAvatarService))

    def test_unregister(self):
        """Testing AvatarServiceHook unregisters services on shutdown"""
        self.assertNotIn(DummyAvatarService, avatar_services)
        AvatarServiceHook(self.extension, DummyAvatarService,
                          start_enabled=True)
        self.assertIn(DummyAvatarService, avatar_services)

        self.extension.shutdown()
        self.assertNotIn(DummyAvatarService, avatar_services)



"""Paginators for iterating over API results.

This is pending deprecation. Consumers should update their imports to use
the classes in :py:mod:`reviewboard.hostingsvcs.base.paginator`.
"""

from reviewboard.hostingsvcs.base.paginator import (APIPaginator,
                                                    BasePaginator,
                                                    InvalidPageError,
                                                    ProxyPaginator)


__all__ = [
    'APIPaginator',
    'BasePaginator',
    'InvalidPageError',
    'ProxyPaginator',
]

__autodoc_excludes__ = __all__



import re

from reviewboard.diffviewer.parser import DiffXParser
from reviewboard.hostingsvcs.errors import HostingServiceError
from reviewboard.scmtools.core import Branch, Commit, ChangeSet
from reviewboard.scmtools.errors import FileNotFoundError, SCMError
from reviewboard.scmtools.git import GitTool


class TestTool(GitTool):
    scmtool_id = 'test'
    name = 'Test'
    diffs_use_absolute_paths = False
    supports_post_commit = True
    supports_history = False

    _PATH_RE = re.compile(
        r'^(?:/(?P<type>data):)?(?P<path>[^;]+)'
        r'(?:;encoding=(?P<encoding>[a-z0-9_-]+))?$')

    def get_repository_info(self):
        return {
            'key1': 'value1',
            'key2': 'value2',
        }

    def get_branches(self):
        return [
            Branch(id='trunk', commit='5', default=True),
            Branch(id='branch1', commit='7', default=False),
        ]

    def get_commits(self, branch=None, start=None):
        if branch == 'bad:hosting-service-error':
            raise HostingServiceError('This is a HostingServiceError')
        elif branch == 'bad:scm-error':
            raise SCMError('This is a SCMError')

        return [
            Commit('user%d' % i, str(i),
                   '2013-01-01T%02d:00:00.0000000' % i,
                   'Commit %d' % i,
                   str(i - 1))
            for i in range(int(start or 10), 0, -1)
        ]

    def get_change(self, commit_id):
        if commit_id == 'bad:hosting-service-error':
            raise HostingServiceError('This is a HostingServiceError')
        elif commit_id == 'bad:scm-error':
            raise SCMError('This is a SCMError')

        return Commit(
            author_name='user1',
            id=commit_id,
            date='2013-01-01T00:00:00.0000000',
            message='Commit summary\n\nCommit description.',
            diff=b'\n'.join([
                b"diff --git a/FILE_FOUND b/FILE_FOUND",
                b"index 712544e4343bf04967eb5ea80257f6c64d6f42c7.."
                b"f88b7f15c03d141d0bb38c8e49bb6c411ebfe1f1 100644",
                b"--- a/FILE_FOUND",
                b"+++ b/FILE_FOUND",
                b"@ -1,1 +1,1 @@",
                b"-blah blah",
                b"+blah",
                b"-",
                b"1.7.1",
            ]))

    def get_file(self, path, revision, **kwargs):
        """Return a file from the repository.

        This testing tool allows for special paths that allow callers to
        optionally define the data to return and the encoding to use for that
        data.

        If the path starts with ``/data:``, then what comes after will be
        returned as data (with a newline appended to the data). Otherwise,
        a standard ``Hello, world!\\n`` will be returned.

        If the path ends with ``;encoding=...``, then whatever is returned will
        be encoded in the specified encoding type.

        If the path starts with ``bad:file-not-found``, then this will simulate
        a :py:class:`reviewboard.scmtools.errors.FileNotFoundError`.

        If the path starts with ``bad:scm-error``, then this will simulate
        a :py:class:`reviewboard.scmtools.errors.SCMError`.

        Version Changed:
            4.0.7:
            Added support for special ``bad:file-not-found`` and
            ``bad:scm-error`` paths.

        Args:
            path (unicode):
                The path to retrieve, optionally with custom data and an
                encoding.

            revision (unicode, unused):
                The revision to retrieve. This is ignored.

            **kwargs (dict, unused):
                Additional keyword arguments for the request.

        Returns:
            bytes:
            The resulting file contents.

        Raises:
            reviewboard.scmtools.errors.FileNotFoundError:
                The path started with ``bad:file-not-found``.

            reviewboard.scmtools.errors.SCMError:
                The path started with ``bad:scm-error``.
        """
        if path.startswith('bad:'):
            if path.startswith('bad:file-not-found'):
                raise FileNotFoundError(path=path,
                                        revision=revision,
                                        **kwargs)
            elif path.startswith('bad:scm-error'):
                raise SCMError('This is a SCMError')

        m = self._PATH_RE.match(path)
        assert m

        path_type = m.group('type')
        path = m.group('path')
        encoding = m.group('encoding') or 'utf-8'

        if path_type == 'data':
            return b'%s\n' % path.encode(encoding)

        return 'Hello, world!\n'.encode(encoding)

    def file_exists(self, path, revision, **kwargs):
        if path == '/FILE_FOUND' or path.startswith('/data:'):
            return True

        return super(TestTool, self).file_exists(path, revision, **kwargs)

    @classmethod
    def check_repository(cls, path, *args, **kwargs):
        pass


class TestToolSupportsPendingChangeSets(TestTool):
    scmtool_id = 'test-supports-pending-changesets'
    name = 'TestToolSupportsPendingChangeSets'
    supports_pending_changesets = True

    def get_changeset(self, changesetid, allow_empty=False):
        changeset = ChangeSet()
        changeset.changenum = changesetid
        changeset.description = 'Hello world!'
        changeset.pending = True
        if not allow_empty:
            changeset.files = ['README.md']
        changeset.summary = 'Added a README markdown to help explain what the'\
            ' repository is used for. Hopefully, this takes off.'
        changeset.testing_done = "None was performed"
        return changeset


class TestToolDiffX(TestTool):
    """SCMTool that uses DiffX diffs.

    Version Added:
        4.0.6
    """

    scmtool_id = 'test-diffx'
    name = 'TestToolDiffX'

    def get_parser(self, data):
        """Return a diff parser used to parse diff data.

        Args:
            data (bytes):
                The diff data to parse.

        Returns:
            reviewboard.diffviewer.diffparser.DiffXParser:
            The diff parser used to parse this data.
        """
        return DiffXParser(data=data)



"""Managers for reviewboard.site.models.

Version Added:
    5.0
"""

from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict
from uuid import uuid4

from django.core.cache import cache
from django.db.models import Count, Manager, Q
from djblets.cache.backend import cache_memoize, make_cache_key
from housekeeping.functions import deprecate_non_keyword_only_args

from reviewboard.deprecation import RemovedInReviewBoard10_0Warning

if TYPE_CHECKING:
    from reviewboard.site.models import AnyOrAllLocalSites, LocalSite


class LocalSiteStatsData(TypedDict):
    """Statistics on Local Sites.

    This includes information on how many Local Sites are in the database,
    total and broken down by public/private access.

    A state UUID is included that will be regenerated every time this data
    is updated. This can be used as a component in other cache keys, ETags,
    or for other unique identifiers in order to help ensure that information
    built on these results are fresh.

    Version Added:
        8.0
    """

    #: The total number of private LocalSites in the database.
    private_count: int

    #: The total number of public LocalSites in the database.
    public_count: int

    #: A UUID specific to the current state generation.
    #:
    #: This can be used to help with other caching and invalidation.
    state_uuid: str

    #: The total number of LocalSites in the database.
    total_count: int


class LocalSiteACLStatsData(TypedDict):
    """Statistics on Local Site access controls.

    This includes information on how many users and administrators are on
    the Local Site and whether the site is public or private.

    A state UUID is included that will be regenerated every time this data
    is updated. This can be used as a component in other cache keys, ETags,
    or for other unique identifiers in order to help ensure that information
    built on these results are fresh.

    Version Added:
        8.0
    """

    #: The number of administrators on the Local Site.
    admin_count: int

    #: Whether the Local Site is public.
    public: bool

    #: A UUID specific to the current state generation.
    #:
    #: This can be used to help with other caching and invalidation.
    state_uuid: str

    #: The number of users on the Local Site.
    user_count: int


class LocalSiteManager(Manager['LocalSite']):
    """Manager for LocalSite models.

    This provides querying and statistics calculations for Local Sites.

    Version Added:
        5.0
    """

    _STATS_CACHE_KEY = 'stats-localsites'

    def get_stats(self) -> LocalSiteStatsData:
        """Return statistics on the LocalSites in the database.

        This will include the total number of :py:class:`LocalSites
        <reviewboard.site.models.LocalSite>`, the number of public LocalSites,
        and the number of private LocalSites.

        This information is cached to avoid querying the database any more
        often than necessary. Statistics will be re-calculated if they fall
        out of cache or if invalidated by saving a LocalSite.

        Version Added:
            5.0

        Returns:
            LocalSiteStatsData:
            The Local Site statistics data.
        """
        def _gen_stats() -> LocalSiteStatsData:
            counts = self.aggregate(
                total=Count('*'),
                public_count=Count('public', filter=Q(public=True)),
            )
            count = counts['total']
            public_count = counts['public_count']
            private_count = count - public_count

            return {
                'private_count': private_count,
                'public_count': public_count,
                'state_uuid': str(uuid4()),
                'total_count': count,
            }

        return cache_memoize(self._STATS_CACHE_KEY, _gen_stats)

    def get_local_site_acl_stats(
        self,
        local_site_or_id: LocalSite | int,
    ) -> LocalSiteACLStatsData | None:
        """Return LocalSite-specific ACL statistics.

        This will include the number of users and administrators on the
        :py:class:`reviewboard.site.models.LocalSite`, along with a UUID
        representing the current membership state.

        This can be used to make determinations on whether to query for
        certain information on a Local Site, and as components in other
        cacheable state.

        This information is cached to avoid querying the database any more
        often than necessary. Statistics will be re-calculated if they fall
        out of cache or if invalidated by altering users or administrators.

        Version Added:
            5.0

        Args:
            local_site_or_id (reviewboard.site.models.LocalSite or int):
                The Local Site instance or ID.

        Returns:
            LocalSiteACLStatsData:
            The Local Site ACL statistics data, or ``None`` if the site
            could not be found.
        """
        def _gen_stats() -> LocalSiteACLStatsData | None:
            model = self.model

            if isinstance(local_site_or_id, int):
                try:
                    local_site = (
                        self.filter(pk=local_site_or_id)
                        .only('public')
                        .get()
                    )
                except model.DoesNotExist:
                    return None
            else:
                local_site = local_site_or_id

            user_count = (
                model.users.through.objects
                .filter(localsite=local_site.pk)
                .count()
            )

            admin_count = (
                model.admins.through.objects
                .filter(localsite=local_site.pk)
                .count()
            )

            return {
                'admin_count': admin_count,
                'public': local_site.public,
                'user_count': user_count,
                'state_uuid': str(uuid4()),
            }

        return cache_memoize(
            self._make_local_site_stats_acl_cache_key(local_site_or_id),
            _gen_stats)

    def has_local_sites(self) -> bool:
        """Return whether there are LocalSites in the database.

        This will optimistically fetch this information from cached
        statistics, only querying the database if necessary.

        Version Added:
            5.0

        Returns:
            bool:
            ``True`` if there are LocalSites. ``False`` if there are not.
        """
        return self.get_stats().get('total_count', 0) > 0

    def has_public_local_sites(self) -> bool:
        """Return whether there are public LocalSites in the database.

        This will optimistically fetch this information from cached
        statistics, only querying the database if necessary.

        Version Added:
            5.0

        Returns:
            bool:
            ``True`` if there are public LocalSites. ``False`` if there are
            not.
        """
        return self.get_stats().get('public_count', 0) > 0

    def has_private_local_sites(self) -> bool:
        """Return whether there are private LocalSites in the database.

        This will optimistically fetch this information from cached
        statistics, only querying the database if necessary.

        Version Added:
            5.0

        Returns:
            bool:
            ``True`` if there are private LocalSites. ``False`` if there are
            not.
        """
        return self.get_stats().get('private_count', 0) > 0

    def invalidate_stats_cache(self) -> None:
        """Invalidate the cache for LocalSite statistics.

        Version Added:
            5.0
        """
        cache.delete(make_cache_key(self._STATS_CACHE_KEY))

    def invalidate_local_site_acl_stats_cache(
        self,
        local_site_or_id: LocalSite | int,
    ) -> None:
        """Invalidate the cache for LocalSite-specific ACL statistics.

        Version Added:
            5.0

        Args:
            local_site_or_id (reviewboard.site.models.LocalSite or int):
                The Local Site instance or ID.
        """
        cache.delete(make_cache_key(
            self._make_local_site_stats_acl_cache_key(local_site_or_id)))

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def build_q(
        self,
        local_site: AnyOrAllLocalSites,
        *,
        allow_all: bool = True,
        local_site_field: str = 'local_site',
    ) -> Q:
        """Return a Q object for matching a Local Site.

        This is used to conditionally build a Q object for filtering by
        Local Site.

        If Local Sites aren't used on the server, or if ``local_site`` is
        :py:attr:`LocalSite.ALL <reviewboard.site.models.LocalSite.ALL>`,
        The Q object will be empty (and will be optimized out of the query).

        Otherwise, a standard Q object matching ``local_site`` will be
        returned.

        This will take care of asserting that compatible arguments are
        provided, so there are no surprises.

        Version Changed:
            8.0:
            ``allow_all`` and ``local_site_field`` must now be provided
            as keyword-only arguments. Support for providing these as
            positional arguments is deprecated and will be removed in
            Review Board 10.

        Version Added:
            5.0

        Args:
            local_site (reviewboard.site.models.LocalSite or
                        reviewboard.site.models.LocalSite.ALL, optional):
                The Local Site value used to filter the queryset, if
                Local Sites are used on the server.

            allow_all (bool, optional):
                Whether :py:attr:`LocalSite.ALL
                <reviewboard.site.models.LocalSite.ALL>` is allowed as a
                value.

            local_site_field (str, optional):
                The name of the field to query. This can be set if a different
                field name or spanned relation (e.g., ``myrel__local_site=``)
                is required.

        Returns:
            django.db.models.Q:
            The resulting query object.

        Raises:
            ValueError:
                Incompatible sets of arguments were provided.
        """
        ALL = self.model.ALL

        if not allow_all and local_site is ALL:
            raise ValueError(
                'allow_all=False was provided, but local_site=LocalSite.ALL '
                'was passed.'
            )

        if local_site is ALL:
            return Q()

        if not self.has_local_sites():
            if local_site not in (None, ALL):
                raise ValueError(
                    f'The server has no LocalSites, but {local_site!r} was '
                    f'passed.'
                )

            return Q()

        return Q(**{local_site_field: local_site})

    def _make_local_site_stats_acl_cache_key(
        self,
        local_site_or_id: LocalSite | int,
    ) -> str:
        """Return a cache key for per-Local Site ACL stats.

        Version Changed:
            8.0:
            This now returns a :py:exc:`ValueError` for an invalid
            ``local_site_or_id`` instead of asserting.

        Args:
            local_site_or_id (reviewboard.site.models.LocalSite or int):
                The Local Site instance or ID.

        Returns:
            str:
            The new cache key.

        Raises:
            ValueError:
                An invalid value for ``local_site_or_id`` was provided.
        """
        if isinstance(local_site_or_id, self.model):
            local_site_id = local_site_or_id.pk
        elif isinstance(local_site_or_id, int):
            local_site_id = local_site_or_id
        else:
            raise ValueError(
                f'Unsupported value provided for local_site_or_id: '
                f'{local_site_or_id!r}'
            )

        return f'local-site-acl-stats-{local_site_id}'



"""Unit tests for reviewboard.search.signal_processor.SignalProcessor."""

import haystack
import kgb
from djblets.siteconfig.models import SiteConfiguration
from haystack.signals import BaseSignalProcessor

from reviewboard.search.signal_processor import SignalProcessor, logger
from reviewboard.testing.testcase import TestCase


class SignalProcessorTests(kgb.SpyAgency, TestCase):
    """Unit tests for reviewboard.search.signal_processor.SignalProcessor."""

    def test_can_process_signals_with_siteconfig(self):
        """Testing SignalProcessor.can_process_signals with stored
        SiteConfiguration
        """
        self.assertIsNotNone(SiteConfiguration.objects.get_current())

        signal_processor = self._create_signal_processor()
        self.assertTrue(signal_processor.can_process_signals)

    def test_can_process_signals_without_siteconfig(self):
        """Testing SignalProcessor.can_process_signals without stored
        SiteConfiguration
        """
        self.spy_on(SiteConfiguration.objects.get_current,
                    op=kgb.SpyOpRaise(SiteConfiguration.DoesNotExist))

        signal_processor = self._create_signal_processor()
        self.assertFalse(signal_processor.can_process_signals)

        # Make sure it works once one has been created.
        SiteConfiguration.objects.get_current.unspy()
        self.assertTrue(signal_processor.can_process_signals)

    def test_handle_delete_with_error(self):
        """Testing SignalProcessor.handle_delete with error"""
        exception = Exception('kaboom!')

        self.spy_on(BaseSignalProcessor.handle_delete,
                    owner=BaseSignalProcessor,
                    op=kgb.SpyOpRaise(exception))
        self.spy_on(logger.error)

        signal_processor = self._create_signal_processor()

        # This should not raise an exception.
        #
        # We'll use some garbage values.
        signal_processor.handle_delete(sender=None,
                                       instance=None)

        self.assertSpyCalled(BaseSignalProcessor.handle_delete)
        self.assertSpyCalledWith(
            logger.error,
            ('Error updating the search index. Check to make sure the '
             'search backend is running and configured correctly, and then '
             'rebuild the search index. Error: %s'),
            exception)

    def test_handle_save_with_error(self):
        """Testing SignalProcessor.handle_save with error"""
        exception = Exception('kaboom!')

        self.spy_on(BaseSignalProcessor.handle_save,
                    owner=BaseSignalProcessor,
                    op=kgb.SpyOpRaise(exception))
        self.spy_on(logger.error)

        signal_processor = self._create_signal_processor()

        # This should not raise an exception.
        #
        # We'll use some garbage values.
        signal_processor.handle_save(sender=None,
                                     instance=None)

        self.assertSpyCalled(BaseSignalProcessor.handle_save)
        self.assertSpyCalledWith(
            logger.error,
            ('Error updating the search index. Check to make sure the '
             'search backend is running and configured correctly, and then '
             'rebuild the search index. Error: %s'),
            exception)

    def _create_signal_processor(self):
        """Return a new instance of our Haystack signal processor.

        Returns:
            reviewboard.search.signal_processor.SignalProcessor:
            The new signal processor.
        """
        return SignalProcessor(haystack.connections,
                               haystack.connection_router)



"""Views for downloading diffs."""

import logging
from typing import Optional

from django.http import Http404, HttpRequest, HttpResponse
from django.shortcuts import get_object_or_404
from django.views.generic.base import View
from djblets.util.http import set_last_modified

from reviewboard.diffviewer.diffutils import (convert_to_unicode,
                                              get_filediff_encodings,
                                              get_original_file,
                                              get_patched_file)
from reviewboard.diffviewer.errors import PatchError
from reviewboard.diffviewer.views import DownloadPatchErrorBundleView
from reviewboard.reviews.views.diff_fragments import ReviewsDiffFragmentView
from reviewboard.reviews.views.mixins import ReviewRequestViewMixin
from reviewboard.scmtools.errors import FileNotFoundError


logger = logging.getLogger(__name__)


class DownloadDiffFileView(ReviewRequestViewMixin, View):
    """Downloads an original or modified file from a diff.

    This will fetch the file from a FileDiff, optionally patching it,
    and return the result as an HttpResponse.
    """

    TYPE_ORIG = 0
    TYPE_MODIFIED = 1

    file_type = TYPE_ORIG

    def get(
        self,
        request: HttpRequest,
        revision: int,
        filediff_id: int,
        *args,
        **kwargs,
    ) -> HttpResponse:
        """Handle HTTP GET requests for this view.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            revision (int):
                The revision of the diff to download the file from.

            filediff_id (int):
                The ID of the FileDiff corresponding to the file to download.

            *args (tuple):
                Positional arguments passed to the handler.

            **kwargs (dict):
                Keyword arguments passed to the handler.

        Returns:
            django.http.HttpResponse:
            The HTTP response to send to the client.
        """
        review_request = self.review_request
        draft = review_request.get_draft(request.user)
        diffset = self.get_diff(revision, draft)
        filediff = get_object_or_404(diffset.files, pk=filediff_id)

        try:
            data = get_original_file(filediff=filediff,
                                     request=request)
        except FileNotFoundError:
            logger.exception(
                'Could not retrieve file "%s" (revision %s) for filediff '
                'ID %s',
                filediff.dest_detail, revision, filediff_id,
                extra={'request': request})
            raise Http404

        if self.file_type == self.TYPE_MODIFIED:
            try:
                data = get_patched_file(source_data=data,
                                        filediff=filediff,
                                        request=request)
            except PatchError:
                logger.exception(
                    'Could not patch file "%s" (revision %s) for filediff '
                    'ID %s',
                    filediff.dest_detail, revision, filediff_id,
                    extra={'request': request})

                raise Http404

        encoding_list = get_filediff_encodings(filediff)
        data = convert_to_unicode(data, encoding_list)[1]

        return HttpResponse(data, content_type='text/plain; charset=utf-8')


class DownloadRawDiffView(ReviewRequestViewMixin, View):
    """View for downloading a raw diff from a review request.

    This will generate a single raw diff file spanning all the FileDiffs
    in a diffset for the revision specified in the URL.
    """

    def get(
        self,
        request: HttpRequest,
        revision: Optional[int] = None,
        *args,
        **kwargs,
    ) -> HttpResponse:
        """Handle HTTP GET requests for this view.

        This will generate the raw diff file and send it to the client.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            revision (int, optional):
                The revision of the diff to download. Defaults to the latest
                revision.

            *args (tuple):
                Positional arguments passed to the handler.

            **kwargs (dict):
                Keyword arguments passed to the handler.

        Returns:
            django.http.HttpResponse:
            The HTTP response to send to the client.
        """
        review_request = self.review_request

        draft = review_request.get_draft(request.user)
        diffset = self.get_diff(revision, draft)

        tool = review_request.repository.get_scmtool()
        data = tool.get_parser(b'').raw_diff(diffset)

        resp = HttpResponse(data, content_type='text/x-patch')

        if diffset.name == 'diff':
            filename = 'rb%d.patch' % review_request.display_id
        else:
            # Get rid of any Unicode characters that may be in the filename.
            filename = diffset.name.encode('ascii', 'ignore').decode('ascii')

            # Content-Disposition headers containing commas break on Chrome 16
            # and newer. To avoid this, replace any commas in the filename with
            # an underscore. Was bug 3704.
            filename = filename.replace(',', '_')

        resp['Content-Disposition'] = 'attachment; filename=%s' % filename
        set_last_modified(resp, diffset.timestamp)

        return resp


class ReviewsDownloadPatchErrorBundleView(DownloadPatchErrorBundleView,
                                          ReviewsDiffFragmentView):
    """A view to download the patch error bundle.

    This view allows users to download a bundle containing data to help debug
    issues when a patch fails to apply. The bundle will contain the diff, the
    original file (as returned by the SCMTool), and the rejects file, if
    applicable.
    """



"""Unit tests for reviewboard.datagrids.columns.ToMeColumn.

Version Added:
    8.0
"""

from __future__ import annotations

from django.utils.safestring import SafeString

from reviewboard.datagrids.columns import ToMeColumn
from reviewboard.datagrids.tests.base import BaseColumnTestCase


class ToMeColumnTests(BaseColumnTestCase):
    """Unit tests for reviewboard.datagrids.columns.ToMeColumn.

    Version Added:
        8.0
    """

    column = ToMeColumn()

    def test_render_data_with_directed_to_user(self) -> None:
        """Testing ToMeColumn.render_data when directed to user"""
        review_request = self.create_review_request(publish=True)

        # This is normally set by the column's augment_queryset().
        self.stateful_column.extra_data['all_to_me'] = {review_request.pk}

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertHTMLEqual(value, '<div title="To Me"><b>\u00bb</b></div>')

    def test_render_data_with_not_directed_to_user(self) -> None:
        """Testing ToMeColumn.render_data when not directed to user"""
        review_request = self.create_review_request(publish=True)

        # This is normally set by the column's augment_queryset().
        self.stateful_column.extra_data['all_to_me'] = set()

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, '')

    def test_to_json_with_directed_to_user(self) -> None:
        """Testing ToMeColumn.to_json when directed to user"""
        review_request = self.create_review_request(publish=True)

        # This is normally set by the column's augment_queryset().
        self.stateful_column.extra_data['all_to_me'] = {review_request.pk}

        self.assertTrue(
            self.column.to_json(self.stateful_column, review_request))

    def test_to_json_with_not_directed_to_user(self) -> None:
        """Testing ToMeColumn.to_json when not directed to user"""
        review_request = self.create_review_request(publish=True)

        # This is normally set by the column's augment_queryset().
        self.stateful_column.extra_data['all_to_me'] = set()

        self.assertFalse(
            self.column.to_json(self.stateful_column, review_request))



"""Review Board e-mail module."""

from django.db.models.signals import post_delete
from djblets.auth.signals import user_registered

from reviewboard.notifications.email.signal_handlers import (
    send_reply_published_mail,
    send_review_published_mail,
    send_review_request_closed_mail,
    send_review_request_published_mail,
    send_user_registered_mail,
    send_webapi_token_created_mail,
    send_webapi_token_deleted_mail,
    send_webapi_token_expired_mail,
    send_webapi_token_updated_mail)
from reviewboard.notifications.email.hooks import (register_email_hook,
                                                   unregister_email_hook)
from reviewboard.reviews.models import ReviewRequest, Review
from reviewboard.reviews.signals import (review_request_published,
                                         review_published, reply_published,
                                         review_request_closed)
from reviewboard.webapi.models import WebAPIToken
from djblets.webapi.signals import (webapi_token_created,
                                    webapi_token_expired,
                                    webapi_token_updated)


def connect_signals() -> None:
    """Connect e-mail callbacks to signals."""
    signal_table = [
        (reply_published, send_reply_published_mail, Review),
        (review_published, send_review_published_mail, Review),
        (review_request_closed, send_review_request_closed_mail,
         ReviewRequest),
        (review_request_published, send_review_request_published_mail,
         ReviewRequest),
        (user_registered, send_user_registered_mail, None),
        (webapi_token_created, send_webapi_token_created_mail, WebAPIToken),
        (webapi_token_expired, send_webapi_token_expired_mail, WebAPIToken),
        (webapi_token_updated, send_webapi_token_updated_mail, WebAPIToken),
        (post_delete, send_webapi_token_deleted_mail, WebAPIToken),
    ]

    for signal, handler, sender in signal_table:
        signal.connect(handler, sender=sender)


__all__ = [
    'register_email_hook',
    'unregister_email_hook',
]



# A light-weight settings file tailored for rbssh.
#
# This is based off the user's primary Review Board settings, but overrides
# the list of installed apps in order to load in the bare minimum settings
# needed for rbssh.

from reviewboard.settings import *  # noqa


# Load only enough to load a SiteConfiguration, needed by rbssh.
#
# NOTE: This is not enough for a full Review Board environment. It's important
#       that SSH backends stay minimal enough to operate within this
#       environment.
INSTALLED_APPS = [
    'django.contrib.contenttypes',
    'django.contrib.sites',
    'djblets.siteconfig',
]



"""API interfaces for Forgejo.

The names of the classes in this module reflect the names used in the Forgejo
swagger UI. The data definitions here do not comprehensively include all
fields, only those that we need for our use.

Version Added:
    8.0
"""

from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel


class APIError(BaseModel):
    """API error for Forgejo.

    Version Added:
        8.0
    """

    message: str
    url: Optional[str]


class AccessToken(BaseModel):
    """API response for a Forgejo access token.

    Version Added:
        8.0
    """

    id: int
    name: str
    scopes: List[str]
    sha1: str
    token_last_eight: str


class GitBlobResponse(BaseModel):
    """API response for a Git blob.

    Version Added:
        8.0
    """

    content: str
    encoding: str
    sha: str
    size: int
    url: str


class CommitMeta(BaseModel):
    """Commit metadata.

    Version Added:
        8.0
    """

    created: str
    sha: str
    url: str


class CommitUser(BaseModel):
    """A user for a commit.

    Version Added:
        8.0
    """

    date: str
    email: str
    name: str


class Commit(BaseModel):
    """API response for a Git commit.

    Version Added:
        8.0
    """

    commit: RepoCommit
    created: str
    parents: List[CommitMeta]
    sha: str


class GitEntry(BaseModel):
    """An entry in a Git tree.

    Version Added:
        8.0
    """

    path: str
    sha: str
    size: int
    type: str


class GitObject(BaseModel):
    """API response for a Git object.

    Version Added:
        8.0
    """

    sha: str
    type: str
    url: str


class GitReference(BaseModel):
    """API data for a Git reference.

    Version Added:
        8.0
    """

    object: GitObject
    ref: str
    url: str


class GitTreeResponse(BaseModel):
    """API data for a Git tree.

    Version Added:
        8.0
    """

    page: int
    sha: str
    total_count: int
    tree: Optional[List[GitEntry]]
    truncated: bool


class Issue(BaseModel):
    """API data for an issue.

    Version Added:
        8.0
    """

    body: str
    id: int
    state: str
    title: str


class Repository(BaseModel):
    """API data for a repository.

    Version Added:
        8.0
    """

    clone_url: str
    default_branch: str
    description: str
    id: int
    name: str
    private: bool


class RepoCommit(BaseModel):
    """Data about a commit.

    Version Added:
        8.0
    """

    author: CommitUser
    committer: CommitUser
    message: str
    tree: CommitMeta



"""Unit tests for reviewboard.attachments.managers.FileAttachmentManager.

Version Added:
    7.0.3:
    This was split off from :py:mod:`reviewboard.attachments.tests`.
"""

from __future__ import annotations

from django.contrib.auth.models import User

from reviewboard.attachments.models import FileAttachment
from reviewboard.attachments.tests.base import BaseFileAttachmentTestCase


class FileAttachmentManagerTests(BaseFileAttachmentTestCase):
    """Unit tests for reviewboard.attachments.managers.FileAttachmentManager.

    Version Added:
        7.0.3:
        This was split off from :py:mod:`reviewboard.attachments.tests`.
    """

    def test_create_from_filediff_sets_relation_counter(self):
        """Testing FileAttachmentManager.create_from_filediff sets
        ReviewRequest.file_attachment_count counter
        """
        user = User.objects.get(username='doc')
        repository = self.create_repository()
        review_request = self.create_review_request(repository=repository,
                                                    submitter=user)
        diffset_history = review_request.diffset_history

        filediff = self.make_filediff(diffset_history=diffset_history)

        self.assertEqual(review_request.file_attachments_count, 0)

        FileAttachment.objects.create_from_filediff(
            filediff,
            file=self.make_uploaded_file(),
            mimetype='image/png')

        review_request.refresh_from_db()
        self.assertEqual(review_request.file_attachments_count, 1)

    def test_create_from_filediff_with_new_and_modified_true(self):
        """Testing FileAttachmentManager.create_from_filediff
        with new FileDiff and modified=True
        """
        filediff = self.make_filediff(is_new=True)
        self.assertTrue(filediff.is_new)

        file_attachment = FileAttachment.objects.create_from_filediff(
            filediff,
            file=self.make_uploaded_file(),
            mimetype='image/png')
        self.assertEqual(file_attachment.repository_id, None)
        self.assertEqual(file_attachment.repo_path, None)
        self.assertEqual(file_attachment.repo_revision, None)
        self.assertEqual(file_attachment.added_in_filediff, filediff)

    def test_create_from_filediff_with_new_and_modified_false(self):
        """Testing FileAttachmentManager.create_from_filediff
        with new FileDiff and modified=False
        """
        filediff = self.make_filediff(is_new=True)
        self.assertTrue(filediff.is_new)

        self.assertRaises(
            AssertionError,
            FileAttachment.objects.create_from_filediff,
            filediff,
            file=self.make_uploaded_file(),
            mimetype='image/png',
            from_modified=False)

    def test_create_from_filediff_with_existing_and_modified_true(self):
        """Testing FileAttachmentManager.create_from_filediff
        with existing FileDiff and modified=True
        """
        filediff = self.make_filediff()
        self.assertFalse(filediff.is_new)

        file_attachment = FileAttachment.objects.create_from_filediff(
            filediff,
            file=self.make_uploaded_file(),
            mimetype='image/png')
        self.assertIsNone(file_attachment.repository_id)
        self.assertIsNone(file_attachment.repo_path)
        self.assertIsNone(file_attachment.repo_revision)
        self.assertEqual(file_attachment.added_in_filediff, filediff)

    def test_create_from_filediff_with_existing_and_modified_false(self):
        """Testing FileAttachmentManager.create_from_filediff
        with existing FileDiff and modified=False
        """
        filediff = self.make_filediff()
        self.assertFalse(filediff.is_new)

        file_attachment = FileAttachment.objects.create_from_filediff(
            filediff,
            file=self.make_uploaded_file(),
            mimetype='image/png',
            from_modified=False)
        self.assertEqual(file_attachment.repository, filediff.get_repository())
        self.assertEqual(file_attachment.repo_path, filediff.source_file)
        self.assertEqual(file_attachment.repo_revision,
                         filediff.source_revision)
        self.assertEqual(file_attachment.added_in_filediff_id, None)

    def test_create_from_filediff_sets_local_site(self):
        """Testing FileAttachmentManager.create_from_filediff sets no local
        site when the filediff's review request has no local site
        """
        uploaded_file = self.make_uploaded_file()

        # Review request with no local site.
        filediff = self.make_filediff()

        file_attachment = FileAttachment.objects.create_from_filediff(
            filediff,
            file=uploaded_file,
            mimetype='image/png')

        self.assertIsNone(file_attachment.local_site)
        self.assertTrue(
            file_attachment.extra_data[self.DEFINED_LOCAL_SITE_KEY])

    def test_create_from_filediff_sets_local_site_with_site(self):
        """Testing FileAttachmentManager.create_from_filediff sets the local
        site from the filediff's review request's local site
        """
        uploaded_file = self.make_uploaded_file()

        # Review request with local site.
        filediff = self.make_filediff(with_local_site=True)
        file_attachment = FileAttachment.objects.create_from_filediff(
            filediff,
            file=uploaded_file,
            mimetype='image/png')

        self.assertEqual(file_attachment.local_site,
                         self.get_local_site(self.local_site_name))
        self.assertTrue(
            file_attachment.extra_data[self.DEFINED_LOCAL_SITE_KEY])

    def test_create_from_filediff_with_local_site(self):
        """Testing FileAttachmentManager.create_from_filediff uses an
        explicitly provided local site
        """
        local_site = self.create_local_site(name='blah')
        filediff = self.make_filediff()

        file_attachment = FileAttachment.objects.create_from_filediff(
            filediff,
            file=self.make_uploaded_file(),
            mimetype='image/png',
            local_site=local_site)

        self.assertEqual(file_attachment.local_site, local_site)
        self.assertTrue(
            file_attachment.extra_data[self.DEFINED_LOCAL_SITE_KEY])

    def test_get_for_filediff_with_new_and_modified_true(self):
        """Testing FileAttachmentManager.get_for_filediff
        with new FileDiff and modified=True
        """
        filediff = self.make_filediff(is_new=True)
        self.assertTrue(filediff.is_new)

        file_attachment = FileAttachment.objects.create_from_filediff(
            filediff,
            file=self.make_uploaded_file(),
            mimetype='image/png')

        self.assertEqual(
            FileAttachment.objects.get_for_filediff(filediff, modified=True),
            file_attachment)

    def test_get_for_filediff_with_new_and_modified_false(self):
        """Testing FileAttachmentManager.get_for_filediff
        with new FileDiff and modified=False
        """
        filediff = self.make_filediff(is_new=True)
        self.assertTrue(filediff.is_new)

        FileAttachment.objects.create_from_filediff(
            filediff,
            file=self.make_uploaded_file(),
            mimetype='image/png')

        self.assertEqual(
            FileAttachment.objects.get_for_filediff(filediff, modified=False),
            None)

    def test_get_for_filediff_with_existing_and_modified_true(self):
        """Testing FileAttachmentManager.get_for_filediff
        with existing FileDiff and modified=True
        """
        filediff = self.make_filediff()
        self.assertFalse(filediff.is_new)

        file_attachment = FileAttachment.objects.create_from_filediff(
            filediff,
            file=self.make_uploaded_file(),
            mimetype='image/png')

        self.assertEqual(
            FileAttachment.objects.get_for_filediff(filediff, modified=True),
            file_attachment)

    def test_get_for_filediff_with_existing_and_modified_false(self):
        """Testing FileAttachmentManager.get_for_filediff
        with existing FileDiff and modified=False
        """
        filediff = self.make_filediff()
        self.assertFalse(filediff.is_new)

        file_attachment = FileAttachment.objects.create_from_filediff(
            filediff,
            file=self.make_uploaded_file(),
            mimetype='image/png',
            from_modified=False)

        self.assertEqual(
            FileAttachment.objects.get_for_filediff(filediff, modified=False),
            file_attachment)



"""Unit tests for reviewboard.actions.base.ActionAttachmentPoint.

Version Added:
    8.0
"""

from __future__ import annotations

from django.template import Context

from reviewboard.actions.base import (ActionPlacement,
                                      ActionAttachmentPoint,
                                      BaseAction)
from reviewboard.actions.renderers import (
    ButtonActionRenderer,
    DetailedMenuActionGroupRenderer,
)
from reviewboard.actions.tests.base import (
    SpecialButtonActionRenderer,
    TestAction,
    TestActionsRegistry,
    TestGroupAction,
    TestGroupItemAction1,
    TestMenuAction,
    TestMenuItemAction,
)
from reviewboard.testing import TestCase


class _MyAction(TestAction):
    placements = [
        ActionPlacement(attachment='test-point'),
    ]


class _MyGroupAction(TestGroupAction):
    placements = [
        ActionPlacement(attachment='test-point'),
    ]


class _MyGroupItemAction1(TestGroupItemAction1):
    label = 'My Group'
    placements = [
        ActionPlacement(attachment='test-point',
                        parent_id=TestGroupAction.action_id),
    ]


class _MyMenuAction(TestMenuAction):
    placements = [
        ActionPlacement(attachment='test-point'),
    ]


class _MyMenuItemAction1(TestMenuItemAction):
    placements = [
        ActionPlacement(attachment='test-point',
                        parent_id=TestMenuAction.action_id),
    ]


class _MyHiddenMenuAction(TestMenuAction):
    action_id = 'hidden-menu-action'
    label = 'Hidden Menu'
    placements = [
        ActionPlacement(attachment='test-point'),
    ]

    def should_render(
        self,
        *,
        context: Context,
    ) -> bool:
        return False


class _MyHiddenMenuItemAction(TestMenuItemAction):
    action_id = 'hidden-menu-item-action'
    label = 'Hidden Menu Item'
    placements = [
        ActionPlacement(attachment='test-point',
                        parent_id='hidden-menu-action'),
    ]


class _MyRegisteredAction(BaseAction):
    action_id = 'new-action'
    label = 'New Action'
    placements = [
        ActionPlacement(attachment='test-point'),
    ]


class _MyActionAttachmentPoint(ActionAttachmentPoint):
    actions = [
        TestAction.action_id,
        TestGroupAction.action_id,
        TestMenuAction.action_id,
    ]
    attachment_point_id = 'test-point'
    default_action_renderer_cls = SpecialButtonActionRenderer
    default_action_group_renderer_cls = DetailedMenuActionGroupRenderer


class ActionAttachmentPointTests(TestCase):
    """Unit tests for ActionAttachmentPoint.

    Version Added:
        8.0
    """

    #: The actions registry used to test against.
    actions_registry: TestActionsRegistry

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()

        actions_registry = TestActionsRegistry()
        actions_registry.register(_MyRegisteredAction())
        actions_registry.register(_MyAction())
        actions_registry.register(_MyGroupAction())
        actions_registry.register(_MyGroupItemAction1())
        actions_registry.register(_MyMenuAction())
        actions_registry.register(_MyMenuItemAction1())
        cls.actions_registry = actions_registry

    @classmethod
    def tearDownClass(cls) -> None:
        cls.actions_registry = None  # type: ignore

        super().tearDownClass()

    def test_init_with_args(self) -> None:
        """Testing ActionAttachmentPoint.__init__ with arguments"""
        attachment_point = ActionAttachmentPoint(
            attachment_point_id='test-point',
            actions=[
                _MyAction.action_id,
                _MyGroupAction.action_id,
            ],
            default_action_renderer_cls=SpecialButtonActionRenderer,
        )

        self.assertEqual(attachment_point.attachment_point_id, 'test-point')
        self.assertEqual(attachment_point.actions, [
            _MyAction.action_id,
            _MyGroupAction.action_id,
        ])
        self.assertIs(attachment_point.default_action_renderer_cls,
                      SpecialButtonActionRenderer)

    def test_init_with_class_attrs(self) -> None:
        """Testing ActionAttachmentPoint.__init__ with class attributes"""
        attachment_point = _MyActionAttachmentPoint()

        self.assertEqual(attachment_point.attachment_point_id, 'test-point')
        self.assertEqual(attachment_point.actions, [
            _MyAction.action_id,
            _MyGroupAction.action_id,
            _MyMenuAction.action_id,
        ])
        self.assertIs(attachment_point.default_action_renderer_cls,
                      SpecialButtonActionRenderer)

    def test_init_with_args_and_class_attrs(self) -> None:
        """Testing ActionAttachmentPoint.__init__ with arguments and class
        attributes
        """
        class MyActionAttachmentPoint(ActionAttachmentPoint):
            attachment_point_id = 'old-test-point'
            actions = [
                _MyGroupAction.action_id,
            ]
            default_action_renderer_cls = ButtonActionRenderer

        attachment_point = MyActionAttachmentPoint(
            attachment_point_id='test-point',
            actions=[
                _MyAction.action_id,
                _MyGroupAction.action_id,
            ],
            default_action_renderer_cls=SpecialButtonActionRenderer,
        )

        self.assertEqual(attachment_point.attachment_point_id, 'test-point')
        self.assertEqual(attachment_point.actions, [
            _MyAction.action_id,
            _MyGroupAction.action_id,
        ])
        self.assertIs(attachment_point.default_action_renderer_cls,
                      SpecialButtonActionRenderer)

    def test_init_with_no_id(self) -> None:
        """Testing ActionAttachmentPoint.__init__ with no ID"""
        message = (
            'attachment_point_id must be provided as an argument or a class '
            'attribute.'
        )

        with self.assertRaisesMessage(AttributeError, message):
            ActionAttachmentPoint()

    def test_iter_actions(self) -> None:
        """Testing ActionAttachmentPoint.iter_actions"""
        actions_registry = self.actions_registry
        attachment_point = _MyActionAttachmentPoint(
            actions_registry=actions_registry,
        )

        self.assertEqual(
            list(attachment_point.iter_actions()),
            [
                actions_registry.get_action(_MyAction.action_id),
                actions_registry.get_action(_MyGroupAction.action_id),
                actions_registry.get_action(_MyMenuAction.action_id),
                actions_registry.get_action(_MyRegisteredAction.action_id),
            ])

    def test_iter_actions_with_include_children(self) -> None:
        """Testing ActionAttachmentPoint.iter_actions with
        include_children=True
        """
        actions_registry = self.actions_registry
        attachment_point = _MyActionAttachmentPoint(
            actions_registry=actions_registry,
        )

        self.assertEqual(
            list(attachment_point.iter_actions(include_children=True)),
            [
                actions_registry.get_action(_MyAction.action_id),
                actions_registry.get_action(_MyGroupAction.action_id),
                actions_registry.get_action(_MyMenuAction.action_id),
                actions_registry.get_action(_MyRegisteredAction.action_id),
                actions_registry.get_action(_MyGroupItemAction1.action_id),
                actions_registry.get_action(_MyMenuItemAction1.action_id),
            ])

    def test_render(self) -> None:
        """Testing ActionAttachmentPoint.render"""
        actions_registry = self.actions_registry
        attachment_point = _MyActionAttachmentPoint(
            actions_registry=actions_registry,
        )

        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        self.assertHTMLEqual(
            attachment_point.render(request=request,
                                    context=context),
            """
            <li class="rb-c-actions__action"
                role="presentation">
             <button aria-label="Test Action 1"
                     class="ink-c-button"
                     id="action-test-point-test"
                     type="button">
              <label class="ink-c-button__label">
               Test Action 1
              </label>
             </button>
            </li>

            <li class="rb-c-actions__action"
                role="menuitem">
             <span id="action-test-point-group-action"
                   role="presentation">
              <a aria-label="Test Group"
                 href="#"
                 role="presentation">
               Test Group
               <span class="ink-i-dropdown"></span>
              </a>
              <div hidden style="display: none;">
               <a id="action-test-point-group-item-1-action"
                  data-custom-rendered="true"
                  href="#"
                  role="menuitem">
                 <h4>
                  My Group
                 </h4>
               </a>
              </div>
             </span>
            </li>

            <li class="rb-c-actions__action"
                role="menuitem">
             <span id="action-test-point-menu-action"
                   role="presentation">
              <a href="#"
                 role="presentation"
                 aria-label="Test Menu">
               Test Menu
               <span class="ink-i-dropdown"/>
              </a>
              <div hidden style="display: none;">
               <a id="action-test-point-menu-item-action"
                  role="button"
                  href="#">
                <span class="my-icon"/>
                Menu Item Action 1
               </a>
              </div>
             </span>
            </li>

            <li class="rb-c-actions__action"
                role="presentation">
             <button aria-label="New Action"
                     class="ink-c-button"
                     id="action-test-point-new-action"
                     type="button">
              <label class="ink-c-button__label">
               New Action
              </label>
             </button>
            </li>
            """)

    def test_render_js(self) -> None:
        """Testing ActionAttachmentPoint.render_js"""
        actions_registry = self.actions_registry
        attachment_point = _MyActionAttachmentPoint(
            actions_registry=actions_registry,
        )

        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        self.assertHTMLEqual(
            attachment_point.render_js(request=request,
                                       context=context),
            """
            page.addActionView(new SpecialButtonActionView({
                "attachmentPointID": "test-point",
                "label": "~~Test Action 1~~",
                "specialKey": [123, 456],
                el: $('#action-test-point-test'),
                model: page.getAction("test"),
            }));

            page.addActionView(new RB.Actions.MenuActionView({
                "attachmentPointID": "test-point",
                el: $('#action-test-point-group-action'),
                model: page.getAction("group-action"),
            }));

            page.addActionView(new RB.Actions.MenuItemActionView({
                "attachmentPointID": "test-point",
                el: $('#action-test-point-group-item-1-action'),
                model: page.getAction("group-item-1-action"),
            }));

            page.addActionView(new RB.Actions.MenuActionView({
                "attachmentPointID": "test-point",
                el: $('#action-test-point-menu-action'),
                model: page.getAction("menu-action"),
            }));

            page.addActionView(new RB.Actions.MenuItemActionView({
                "attachmentPointID": "test-point",
                el: $('#action-test-point-menu-item-action'),
                model: page.getAction("menu-item-action"),
            }));

            page.addActionView(new SpecialButtonActionView({
                "attachmentPointID": "test-point",
                "label": "~~New Action~~",
                "specialKey": [123, 456],
                el: $('#action-test-point-new-action'),
                model: page.getAction("new-action"),
            }));
            """)

    def test_render_js_with_hidden_parent(self) -> None:
        """Testing ActionAttachmentPoint.render_js skips children when parent
        should_render is False
        """
        actions_registry = TestActionsRegistry()
        actions_registry.register(_MyHiddenMenuAction())
        actions_registry.register(_MyHiddenMenuItemAction())

        attachment_point = _MyActionAttachmentPoint(
            actions=[
                _MyHiddenMenuAction.action_id,
            ],
            actions_registry=actions_registry,
        )

        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        # Both the parent and children should be skipped, resulting in an empty
        # string.
        self.assertHTMLEqual(
            attachment_point.render_js(request=request,
                                       context=context),
            '')
